# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Autopkgtest work request plugin."""

import textwrap
from typing import ClassVar, override

from django.http import HttpResponseBase
from django.template.loader import get_template
from django.template.response import SimpleTemplateResponse

from debusine.artifacts import AutopkgtestArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianAutopkgtest,
    DebianAutopkgtestSource,
)
from debusine.db.models import Artifact, WorkRequest, Workspace
from debusine.db.playground import scenarios
from debusine.tasks.models import (
    AutopkgtestData,
    AutopkgtestDynamicData,
    AutopkgtestInput,
    BackendType,
    LookupMultiple,
    LookupSingle,
)
from debusine.test.django import TestCase
from debusine.web.views.autopkgtest import (
    AutopkgtestLogParser,
    AutopkgtestTab,
    AutopkgtestViewWorkRequestPlugin,
)
from debusine.web.views.files import LogFileWidget
from debusine.web.views.work_request import Tab


class AutopkgtestViewWorkRequestPluginTests(TestCase):
    """Tests for AutopkgtestViewWorkRequestPlugin."""

    environment_artifact: ClassVar[Artifact]
    source_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.environment_artifact, _ = cls.playground.create_artifact(
            paths=["system.tar.xz"],
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={
                "vendor": "debian",
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": "",
            },
            create_files=True,
        )
        cls.source_artifact, _ = cls.playground.create_artifact(
            paths=["hello.dsc"],
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={"name": "hello", "version": "1.0", "type": "dpkg"},
            create_files=True,
        )

    def get_tab(self, work_request: WorkRequest) -> AutopkgtestTab:
        """Return the tab for a work request or a plugin."""
        plugin = AutopkgtestViewWorkRequestPlugin(work_request)
        tabs: list[Tab] = []
        plugin.add_tabs(tabs)
        self.assertEqual(len(tabs), 1)
        tab = tabs[0]
        assert isinstance(tab, AutopkgtestTab)
        self.assertEqual(tab.name, "autopkgtest")
        self.assertEqual(tab.label, "Autopkgtest")
        return tab

    def create_work_request_for_artifacts(
        self,
        environment_artifact: Artifact,
        source_artifact: Artifact,
        binary_artifacts: list[Artifact],
        context_artifacts: list[Artifact] | None = None,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create an autopkgtest work request for some artifacts."""
        environment_lookup: int | str
        if workspace is None:
            workspace = self.playground.get_default_workspace()
        if "vendor" in environment_artifact.data:
            environments_collection = self.playground.create_collection(
                environment_artifact.data["vendor"],
                CollectionCategory.ENVIRONMENTS,
                workspace=workspace,
            )
            environments_collection.manager.add_artifact(
                environment_artifact,
                user=self.playground.get_default_user(),
                variables={"backend": BackendType.UNSHARE},
            )
            environment_lookup = "{vendor}/match:codename={codename}".format(
                **environment_artifact.data
            )
        else:
            environment_lookup = environment_artifact.id
        task_data_input = AutopkgtestInput(
            source_artifact=source_artifact.id,
            binary_artifacts=LookupMultiple.model_validate(
                [binary_artifact.id for binary_artifact in binary_artifacts]
            ),
        )
        if context_artifacts:
            task_data_input.context_artifacts = LookupMultiple.model_validate(
                [context_artifact.id for context_artifact in context_artifacts]
            )

        return self.playground.create_worker_task(
            result=WorkRequest.Results.SUCCESS,
            task_data=AutopkgtestData(
                build_architecture="amd64",
                environment=environment_lookup,
                input=task_data_input,
            ),
            task_name="autopkgtest",
            workspace=workspace,
        )

    def create_autopkgtest_artifact(
        self, work_request: WorkRequest
    ) -> Artifact:
        """Create an AutopkgtestArtifact for a work request."""
        artifact, _ = self.playground.create_artifact(
            paths=["log", "summary"],
            work_request=work_request,
            create_files=True,
        )
        artifact.category = AutopkgtestArtifact._category
        artifact.data = {
            "results": {
                "command1": {"status": "PASS"},
                "command2": {"status": "FAIL", "details": "partial"},
            },
            "cmdline": "/usr/bin/hello",
            "source_package": {
                "name": "hello",
                "version": "1.0",
                "url": "https://example.org/hello.deb",
            },
            "architecture": "amd64",
            "distribution": "debian:sid",
        }
        artifact.save()
        return artifact

    def test_is_enabled(self) -> None:
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        autopkgtest_view = AutopkgtestViewWorkRequestPlugin(work_request)
        self.assertTrue(autopkgtest_view.is_enabled())

    def test_get_context_data_simple(self) -> None:
        """get_context_data returns the expected data for a simple case."""
        binary_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-1.deb"], create_files=True
        )
        binary_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact,
            self.source_artifact,
            [binary_artifact_1, binary_artifact_2],
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        result_artifact = self.create_autopkgtest_artifact(work_request)
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data,
            {
                "request_data": {
                    "backend": BackendType.AUTO,
                    "binary_artifacts": {
                        "lookup": [binary_artifact_1.id, binary_artifact_2.id],
                        "artifacts": [
                            {
                                "files": ["hello-1.deb"],
                                "id": binary_artifact_1.id,
                                "artifact": binary_artifact_1,
                            },
                            {
                                "files": ["hello-2.deb"],
                                "id": binary_artifact_2.id,
                                "artifact": binary_artifact_2,
                            },
                        ],
                    },
                    "build_architecture": "amd64",
                    "codename": "bookworm",
                    "context_artifacts": {"lookup": [], "artifacts": []},
                    "debug_level": 0,
                    "exclude_tests": [],
                    "extra_repositories": None,
                    "extra_environment": {},
                    "fail_on_scenarios": ["failed"],
                    "include_tests": [],
                    "needs_internet": "run",
                    "source_artifact": {
                        "files": ["hello.dsc"],
                        "id": self.source_artifact.id,
                        "lookup": self.source_artifact.id,
                        "artifact": self.source_artifact,
                    },
                    "source_name": "hello",
                    "source_version": "1.0",
                    "task_description": "hello_1.0_amd64 in debian:bookworm",
                    "timeout": None,
                    "use_packages_from_base_repository": False,
                    "vendor": "debian",
                },
                "result": work_request.result,
                "result_artifact": result_artifact,
                "result_data": [
                    ("command1", "PASS", None),
                    ("command2", "FAIL", "partial"),
                ],
            },
        )

    def test_get_context_data_with_context_artifacts(self) -> None:
        """Any context artifacts show up in the context data."""
        context_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-rdep-1.deb"], create_files=True
        )
        context_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-rdep-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact,
            self.source_artifact,
            [],
            context_artifacts=[context_artifact_1, context_artifact_2],
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
            input_context_artifacts_ids=[
                context_artifact_1.id,
                context_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()

        self.assertEqual(
            context_data["request_data"]["context_artifacts"],
            {
                "lookup": [context_artifact_1.id, context_artifact_2.id],
                "artifacts": [
                    {
                        "files": ["hello-rdep-1.deb"],
                        "id": context_artifact_1.id,
                    },
                    {
                        "files": ["hello-rdep-2.deb"],
                        "id": context_artifact_2.id,
                    },
                ],
            },
        )

    def test_context_data_no_artifacts(self) -> None:
        """If there are no output artifacts, context data omits that info."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        context_data = self.get_tab(work_request).get_context_data()
        self.assertCountEqual(context_data.keys(), {"request_data", "result"})

    def test_context_data_no_dynamic_task_data(self) -> None:
        """If there is no dynamic data, context data omits artifact info."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"],
            "UNKNOWN_UNKNOWN_amd64",
        )
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {
                "lookup": self.source_artifact.id,
                "id": None,
                "files": [],
                "artifact": None,
            },
        )

    def test_context_data_missing_environment_artifact(self) -> None:
        """If the environment artifact is missing, context data omits it."""
        environment_artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        environment_artifact.delete()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"], "hello_1.0_amd64"
        )

    def test_context_data_environment_artifact_wrong_category(self) -> None:
        """We cope with the environment artifact having an odd category."""
        environment_artifact, _ = self.playground.create_artifact(
            paths=["system.tar.xz"],
            category=ArtifactCategory.TEST,
            create_files=True,
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        tab = self.get_tab(work_request)
        context_data = tab.get_context_data()
        self.assertIsNone(context_data["request_data"]["codename"])
        self.assertIsNone(context_data["request_data"]["vendor"])
        self.assertEqual(
            context_data["request_data"]["task_description"], "hello_1.0_amd64"
        )

    def test_context_data_missing_source_artifact(self) -> None:
        """If the source artifact is missing, context data omits its files."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        source_artifact_id = self.source_artifact.id
        self.source_artifact.files.clear()
        self.source_artifact.delete()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"],
            {
                "lookup": source_artifact_id,
                "id": source_artifact_id,
                "artifact": None,
                "files": [],
            },
        )

    def test_context_data_source_artifact_wrong_category(self) -> None:
        """We cope with the source artifact having an odd category."""
        self.source_artifact.category = ArtifactCategory.TEST
        self.source_artifact.save()
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertIsNone(context_data["request_data"]["source_name"])
        self.assertIsNone(context_data["request_data"]["source_version"])
        self.assertEqual(
            context_data["request_data"]["task_description"],
            "UNKNOWN_UNKNOWN_amd64 in debian:bookworm",
        )

    def test_context_data_missing_binary_artifact(self) -> None:
        """If a binary artifact is missing, context data omits its files."""
        binary_artifact_1, _ = self.playground.create_artifact(
            paths=["hello-1.deb"], create_files=True
        )
        binary_artifact_2, _ = self.playground.create_artifact(
            paths=["hello-2.deb"], create_files=True
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=self.source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        binary_artifact_1_id = binary_artifact_1.id
        binary_artifact_1.files.clear()
        binary_artifact_1.delete()
        tab = self.get_tab(work_request)
        context_data = tab.get_context_data()
        self.assertEqual(
            context_data["request_data"]["binary_artifacts"],
            {
                "lookup": work_request.task_data["input"]["binary_artifacts"],
                "artifacts": [
                    {
                        "id": binary_artifact_1_id,
                        "files": [],
                        "artifact": None,
                    },
                    {
                        "id": binary_artifact_2.id,
                        "files": ["hello-2.deb"],
                        "artifact": binary_artifact_2,
                    },
                ],
            },
        )

    def test_get_context_data_skips_incomplete_files(self) -> None:
        """get_context_data() skips incomplete files."""
        source_artifact, _ = self.playground.create_artifact(
            paths=["hello.dsc"],
            category=ArtifactCategory.SOURCE_PACKAGE,
            data={"name": "hello", "version": "1.0", "type": "dpkg"},
            create_files=True,
            skip_add_files_in_store=True,
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, source_artifact, []
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=self.environment_artifact.id,
            input_source_artifact_id=source_artifact.id,
            input_binary_artifacts_ids=[],
        ).model_dump()
        work_request.save()

        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["source_artifact"]["files"], []
        )

    def assert_contains_artifact_information(
        self,
        response: HttpResponseBase,
        description: str,
        artifact_lookup: LookupSingle,
        artifact_id: int,
    ) -> None:
        """Assert contains information of an artifact: link and files."""
        artifact = Artifact.objects.get(id=artifact_id)
        files = artifact.fileinartifact_set.order_by("path").values_list(
            "path", flat=True
        )
        artifact_path = artifact.get_absolute_url()
        artifact_link = f'<a href="{artifact_path}">#{artifact_id}</a>'
        files_li = "".join([f"<li>{file}</li>" for file in files])
        self.assertContains(
            response,
            f"<li>{description} ({artifact_lookup}: {artifact_link})"
            f"<ul>{files_li}</ul>"
            f"</li>",
            html=True,
        )

    def assert_contains_artifacts_information(
        self,
        response: HttpResponseBase,
        description: str,
        artifact_lookup: LookupMultiple,
        artifact_ids: list[int],
    ) -> None:
        """Assert contains information of a set of artifacts."""
        expected_response = (
            f"<li>{description} ({artifact_lookup.export()})<ul>"
        )
        for artifact_id in artifact_ids:
            artifact = Artifact.objects.get(id=artifact_id)
            files = artifact.fileinartifact_set.order_by("path").values_list(
                "path", flat=True
            )
            artifact_path = artifact.get_absolute_url()
            artifact_link = f'<a href="{artifact_path}">#{artifact_id}</a>'
            files_li = "".join([f"<li>{file}</li>" for file in files])
            expected_response += f"<li>{artifact_link}<ul>{files_li}</ul></li>"
        expected_response += "</ul></li>"
        self.assertContains(response, expected_response, html=True)

    def test_template_output(self) -> None:
        """Generic output of the template."""
        workspace = self.playground.create_workspace(name="Public", public=True)
        environment = self.playground.create_debian_environment(
            workspace=workspace
        )
        assert environment.artifact
        source_artifact = self.playground.create_source_artifact(
            workspace=workspace
        )
        binary_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello-1"
            )
        )
        binary_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello-2"
            )
        )
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            environment.artifact,
            source_artifact,
            [binary_artifact_1, binary_artifact_2],
            workspace=workspace,
        )
        work_request.dynamic_task_data = AutopkgtestDynamicData(
            environment_id=environment.artifact.id,
            input_source_artifact_id=source_artifact.id,
            input_binary_artifacts_ids=[
                binary_artifact_1.id,
                binary_artifact_2.id,
            ],
        ).model_dump()
        work_request.save()
        self.create_autopkgtest_artifact(work_request)

        response = self.client.get(work_request.get_absolute_url())
        tree = self.assertResponseHTML(response, dump_on_error=True)
        title = self.assertHasElement(tree, "//head/title")
        self.assertTextContentEqual(
            title, f"debusine - Public - #{work_request.id}: autopkgtest"
        )

        result_output = get_template("web/_work_request-result.html").render(
            {"result": work_request.result}
        )
        self.assertContains(response, f"Result: {result_output}", html=True)

        self.assert_contains_artifact_information(
            response, "Source artifact", source_artifact.id, source_artifact.id
        )

        self.assert_contains_artifacts_information(
            response,
            "Binary artifacts",
            LookupMultiple((binary_artifact_1.id, binary_artifact_2.id)),
            [binary_artifact_1.id, binary_artifact_2.id],
        )

    def test_get_fail_on_scenarios(self) -> None:
        """_get_fail_on_scenarios processes the task's fail_on correctly."""
        sub_tests = [
            ({"failed_test": True}, ["failed"]),
            (
                {
                    "failed_test": False,
                    "flaky_test": True,
                    "skipped_test": False,
                },
                ["flaky"],
            ),
            (
                {
                    "failed_test": True,
                    "flaky_test": True,
                    "skipped_test": True,
                },
                ["failed", "flaky", "skipped"],
            ),
        ]

        for fail_on, expected_scenarios in sub_tests:
            with self.subTest(fail_on=fail_on):
                self.assertEqual(
                    AutopkgtestTab._get_fail_on_scenarios(fail_on),
                    expected_scenarios,
                )

    def test_show_configured_task_data(self) -> None:
        """The view shows configured task data if present."""
        work_request: WorkRequest = self.create_work_request_for_artifacts(
            self.environment_artifact, self.source_artifact, []
        )

        # Try unconfigured
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["build_architecture"], "amd64"
        )

        # Try configured
        work_request.configured_task_data = work_request.task_data.copy()
        work_request.configured_task_data["build_architecture"] = "arm64"
        work_request.save()
        context_data = self.get_tab(work_request).get_context_data()
        self.assertEqual(
            context_data["request_data"]["build_architecture"], "arm64"
        )


class AutopkgtestLogParserTests(TestCase):
    """Tests for AutopkgtestLogParser class."""

    scenario = scenarios.PackageBuildLog()

    def test_parse_log(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        sections, number_of_lines = AutopkgtestLogParser(
            self.scenario.autopkgtest_log
        ).parse()
        self.assertEqual(
            number_of_lines, self.scenario.autopkgtest_log.count(b"\n")
        )

        self.assertEqual(
            sections,
            [
                Section(
                    text="Preparation",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="start run",
                            color=None,
                            subsections=[],
                            line=1,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                        SubSection(
                            text="test bed setup",
                            color=None,
                            subsections=[],
                            line=6,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                        SubSection(
                            text="apt-source",
                            color=None,
                            subsections=[],
                            line=10,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                    ],
                ),
                Section(
                    text="Tests",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="preparing testbed pybuild-autopkgtest",
                            color=None,
                            subsections=[],
                            line=15,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip=None,
                        ),
                        SubSection(
                            text="pybuild-autopkgtest",
                            color="success",
                            subsections=[],
                            line=23,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip="Pass",
                        ),
                        SubSection(
                            text="preparing testbed test-suite",
                            color=None,
                            subsections=[],
                            line=31,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip=None,
                        ),
                        SubSection(
                            text="test-suite",
                            color="danger",
                            subsections=[],
                            line=43,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip="Fail",
                        ),
                        SubSection(
                            text="test-suite",
                            color=None,
                            subsections=[],
                            line=49,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        ),
                        SubSection(
                            text="preparing testbed test-suite-skip",
                            color="secondary",
                            subsections=[],
                            line=50,
                            result=None,
                            expanded=False,
                            include_in_table_of_contents=False,
                            tooltip="Skip",
                        ),
                    ],
                ),
                Section(
                    text="Closing",
                    color="primary",
                    subsections=[
                        SubSection(
                            text="summary",
                            color=None,
                            subsections=[],
                            line=53,
                            result=None,
                            expanded=True,
                            include_in_table_of_contents=True,
                            tooltip=None,
                        )
                    ],
                ),
            ],
        )

    def test_parse_log_without_apt_source(self) -> None:
        Section = AutopkgtestLogParser.Section
        SubSection = AutopkgtestLogParser.SubSection
        Section.model_rebuild()
        SubSection.model_rebuild()

        log_contents = textwrap.dedent(
            """\
        line
        line
        15s autopkgtest [22:36:01]: test m-a-autopkgtest: preparing testbed
        line
        43s autopkgtest [22:36:29]: test m-a-autopkgtest: [-----------------------
        44s I: No Linux header packages are installed.
        line
        line
        """  # noqa: E501
        ).encode()

        sections, number_of_lines = AutopkgtestLogParser(log_contents).parse()
        self.assertEqual(number_of_lines, log_contents.count(b"\n"))
        self.assertEqual(len(sections), 3)
        self.assertEqual(sections[0].subsections, [])

        # 1 subsection: preparing and the test itself
        self.assertEqual(
            sections[1].subsections,
            [
                SubSection(
                    text="preparing testbed m-a-autopkgtest",
                    color=None,
                    subsections=[],
                    line=3,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=False,
                    tooltip=None,
                ),
                SubSection(
                    text="m-a-autopkgtest",
                    color=None,
                    subsections=[],
                    line=5,
                    result=None,
                    expanded=False,
                    include_in_table_of_contents=True,
                    tooltip=None,
                ),
            ],
        )
        self.assertEqual(sections[2].subsections, [])


class AutopkgtestArtifactPluginTests(TestCase):
    """Tests for AutopkgtestArtifactPlugin class."""

    scenario = scenarios.PackageBuildLog()

    def test_get_context_data(self) -> None:
        artifact = self.scenario.autopkgtest_log_artifact
        log_file = artifact.fileinartifact_set.get(path="log")
        log_file.content_type = "text/plain; charset=utf-8"
        log_file.save()

        response = self.client.get(artifact.get_absolute_url())

        tree = self.assertResponseHTML(response)

        assert isinstance(response, SimpleTemplateResponse)
        assert response.context_data is not None
        self.assertEqual(
            response.context_data["specialized_tab"],
            {
                "label": "Autopkgtest log",
                "slug": "autopkgtest-log",
                "template": "web/_autopkgtest-artifact-detail.html",
            },
        )

        (sections, _) = AutopkgtestLogParser(
            self.scenario.autopkgtest_log
        ).parse()

        self.assertEqual(response.context_data["sections"], sections)
        FileSection = LogFileWidget.Section

        self.assertEqual(
            response.context_data["log_widget"]._sections,
            [
                FileSection(
                    title="start run",
                    start_line=1,
                    end_line=5,
                    expanded=True,
                ),
                FileSection(
                    title="test bed setup",
                    start_line=6,
                    end_line=9,
                    expanded=False,
                ),
                FileSection(
                    title="apt-source",
                    start_line=10,
                    end_line=14,
                    expanded=False,
                ),
                FileSection(
                    title="preparing testbed pybuild-autopkgtest",
                    start_line=15,
                    end_line=22,
                    expanded=False,
                ),
                FileSection(
                    title="pybuild-autopkgtest",
                    start_line=23,
                    end_line=30,
                    expanded=False,
                ),
                FileSection(
                    title="preparing testbed test-suite",
                    start_line=31,
                    end_line=42,
                    expanded=False,
                ),
                FileSection(
                    title="test-suite",
                    start_line=43,
                    end_line=48,
                    expanded=True,
                ),
                FileSection(
                    title="test-suite",
                    start_line=49,
                    end_line=49,
                    expanded=False,
                ),
                FileSection(
                    title="preparing testbed test-suite-skip",
                    start_line=50,
                    end_line=52,
                    expanded=False,
                ),
                FileSection(
                    title="summary",
                    start_line=53,
                    end_line=57,
                    expanded=True,
                ),
            ],
        )

        # Assert line number rendering is correct (has the ID and content)
        line_number_1 = self.assertHasElement(tree, "//a[@id='L1']")
        self.assertTextContentEqual(line_number_1, "1")

    def test_get_context_data_empty_log(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=True,
            data=data,
        )
        log_file = artifact.fileinartifact_set.get(path="log")
        log_file.content_type = "text/plain; charset=utf-8"
        log_file.save()

        response = self.client.get(artifact.get_absolute_url())

        self.assertResponseHTML(response)

    def test_autopkgtest_without_log(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=False,
            data=data,
        )

        response = self.client.get(artifact.get_absolute_url())
        tree = self.assertResponseHTML(response)

        error_paragraph = self.assertHasElement(
            tree, "//div[@id='error-artifact']"
        )

        self.assertTextContentEqual(
            error_paragraph,
            "Cannot display autopkgtest: Artifact does not contain log",
        )

    def test_autopkgtest_with_file_in_artifact_error(self) -> None:
        data = DebianAutopkgtest(
            results={},
            cmdline="autopkgtest",
            source_package=DebianAutopkgtestSource(
                name="hello", version="1.0", url="https://example.com"
            ),
            architecture="amd64",
            distribution="debian:trixie",
        )

        artifact, _ = self.playground.create_artifact(
            paths={"log": b""},
            category=ArtifactCategory.AUTOPKGTEST,
            create_files=True,
            skip_add_files_in_store=True,
            data=data,
        )

        response = self.client.get(artifact.get_absolute_url())
        tree = self.assertResponseHTML(response)

        error = self.assertHasElement(
            tree,
            "//div[@id='error-artifact']",
        )

        self.assertTextContentEqual(
            error,
            'Cannot display autopkgtest: '
            'File "log" has not completed uploading.',
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Empty Workflow, used for testing."""

from typing import override

from debusine.db.models.tasks import DefaultDynamicData
from debusine.server.workflows.base import Workflow
from debusine.server.workflows.models import BaseWorkflowData
from debusine.tasks.models import BaseDynamicTaskData


class NoopWorkflow(
    Workflow[BaseWorkflowData, BaseDynamicTaskData],
    DefaultDynamicData[BaseWorkflowData],
):
    """Workflow that does nothing."""

    TASK_NAME = "noop"

    @override
    def populate(self) -> None:
        """Do nothing."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for base workflow code."""

import logging
from typing import ClassVar
from unittest import mock

from debusine.client.models import RuntimeParameter
from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.workflows.base import (
    Workflow,
    WorkflowValidationError,
    orchestrate_workflow,
)
from debusine.server.workflows.models import (
    BaseWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import SampleWorkflow
from debusine.tasks.models import (
    BaseDynamicTaskData,
    OutputData,
    OutputDataError,
)
from debusine.tasks.server import TaskDatabaseInterface
from debusine.test.django import TestCase, preserve_db_task_registry


class OrchestrateWorkflowTests(TestCase):
    """Test orchestrate_workflow()."""

    def test_workflow_callback_success(self) -> None:
        """A workflow callback is run and marked as completed."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=True,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow_callback_failure(self) -> None:
        """A workflow callback that returns False is marked as failed."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=False,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.FAILURE)

    def test_workflow_callback_step_not_implemented(self) -> None:
        """A workflow callback must have a matching ``callback_*`` method."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = (
            "Orchestrator failed: Unhandled workflow callback step: test"
        )

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback_other",
                create=True,
                return_value=True,
            ) as mock_noop_callback_other,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback_other.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    @preserve_db_task_registry()
    def test_workflow_callback_computes_dynamic_data(self) -> None:
        """Dynamic task data is computed for workflow callbacks."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            def compute_dynamic_data(
                self,
                task_database: TaskDatabaseInterface,  # noqa: ARG002, U100
            ) -> ExampleDynamicData:
                return ExampleDynamicData(dynamic="foo")

            def populate(self) -> None:
                """Unused abstract method from Workflow."""
                raise NotImplementedError()

        parent = self.playground.create_workflow(
            task_name="example", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch.object(ExampleWorkflow, "callback") as mock_callback:
            self.assertTrue(orchestrate_workflow(wr))

        mock_callback.assert_called_once_with(wr)
        parent.refresh_from_db()
        self.assertEqual(parent.dynamic_task_data, {"dynamic": "foo"})
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    @preserve_db_task_registry()
    def test_workflow_callback_computes_dynamic_data_only_once(self) -> None:
        """If a workflow already has dynamic data, it is left alone."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str = "unset"

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            def populate(self) -> None:
                """Unused abstract method from Workflow."""
                raise NotImplementedError()

        WorkRequest.objects.all().delete()
        parent = self.playground.create_workflow(task_name="example")
        parent.dynamic_task_data = {"dynamic": "foo"}
        parent.save()
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with (
            mock.patch.object(
                ExampleWorkflow, "compute_dynamic_data"
            ) as mock_compute_dynamic_data,
            mock.patch.object(ExampleWorkflow, "callback") as mock_callback,
        ):
            self.assertTrue(orchestrate_workflow(wr))

        mock_compute_dynamic_data.assert_not_called()
        mock_callback.assert_called_once_with(wr)
        parent.refresh_from_db()
        self.assertEqual(parent.dynamic_task_data, {"dynamic": "foo"})
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow_callback_outside_workflow(self) -> None:
        """A workflow callback outside a workflow is skipped."""
        wr = self.playground.create_internal_task(
            task_name="workflow", mark_pending=True
        )
        expected_message = "Workflow callback is not contained in a workflow"

        with self.assertLogsContains(
            f"Error running work request Internal/workflow ({wr.id}): "
            f"{expected_message}",
            logger="debusine.server.workflows.base",
            level=logging.WARNING,
        ):
            self.assertFalse(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="configure-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_bad_task_data(self) -> None:
        """A workflow callback with bad task data is skipped."""
        parent = self.playground.create_workflow(
            task_name="noop", task_data={"nonsense": ""}, validate=False
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Failed to configure: 1 validation error"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback"
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        assert isinstance(wr.output_data, OutputData)
        assert wr.output_data.errors is not None
        self.assertEqual(len(wr.output_data.errors), 1)
        self.assertTrue(
            wr.output_data.errors[0].message.startswith(expected_message)
        )
        self.assertEqual(wr.output_data.errors[0].code, "configure-failed")

    def test_workflow_callback_with_failing_compute_dynamic_data(self) -> None:
        """A workflow callback where computing dynamic data fails is skipped."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = (
            f"Failed to compute dynamic data for Workflow/noop ({parent.id}): "
            f"Boom"
        )

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow"
                ".compute_dynamic_data",
                side_effect=Exception("Boom"),
            ),
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback"
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="dynamic-data-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_fails(self) -> None:
        """A workflow callback that fails is logged."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Orchestrator failed: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback",
                side_effect=ValueError("Boom"),
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_called_once_with(wr)
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_fails_without_context(self) -> None:
        """A callback failure by bare exception logs the exception name."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Orchestrator failed: AssertionError()"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback",
                side_effect=AssertionError(),
            ),
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow(self) -> None:
        """A workflow is populated and left running."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def populate() -> None:
            wr.create_child_worker("noop")

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=populate,
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)

    def test_workflow_empty(self) -> None:
        """An empty workflow is populated and marked as completed."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate"
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    @preserve_db_task_registry()
    def test_workflow_empty_sub_workflow(self) -> None:
        """A workflow with an empty sub-workflow isn't prematurely completed."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            TASK_NAME = "example"

            def populate(self) -> None:
                noop1 = self.work_request.create_child_workflow("noop")
                self.orchestrate_child(noop1)

                # Normally we'd orchestrate this sub-workflow as well, but
                # we skip that for the purposes of this test.
                self.work_request.create_child_workflow("noop")

        wr = self.playground.create_workflow(
            task_name="example", mark_running=True
        )

        self.assertTrue(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertQuerySetEqual(
            wr.children.order_by("id").values_list("status", "result"),
            [
                (WorkRequest.Statuses.COMPLETED, WorkRequest.Results.SUCCESS),
                (WorkRequest.Statuses.PENDING, WorkRequest.Results.NONE),
            ],
        )

    @preserve_db_task_registry()
    def test_workflow_computes_dynamic_data(self) -> None:
        """Dynamic task data is computed for workflows."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            def compute_dynamic_data(
                self,
                task_database: TaskDatabaseInterface,  # noqa: ARG002, U100
            ) -> ExampleDynamicData:
                return ExampleDynamicData(dynamic="foo")

            def populate(self) -> None:
                self.work_request.create_child_worker("noop")

        wr = self.playground.create_workflow(
            task_name="example", mark_running=True
        )

        self.assertTrue(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.dynamic_task_data, {"dynamic": "foo"})
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)

    def test_workflow_bad_task_data(self) -> None:
        """A workflow with bad task data is skipped."""
        # Bad task data would normally be caught by create_workflow.  Force
        # it to happen here by changing the task data after initial
        # creation.  (A more realistic case might be one where the
        # definition of a workflow changes but existing data isn't
        # migrated.)
        wr = self.playground.create_workflow(task_name="noop")
        wr.task_data = {"nonsense": ""}
        wr.configured_task_data = {"nonsense": ""}
        wr.save()
        self.playground.advance_work_request(wr, mark_running=True)
        expected_message = "Failed to configure: 1 validation error"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        assert isinstance(wr.output_data, OutputData)
        assert wr.output_data.errors is not None
        self.assertEqual(len(wr.output_data.errors), 1)
        self.assertTrue(
            wr.output_data.errors[0].message.startswith(expected_message)
        )
        self.assertEqual(wr.output_data.errors[0].code, "configure-failed")

    def test_workflows_with_failing_compute_dynamic_data(self) -> None:
        """A workflow where computing dynamic data fails is skipped."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        expected_message = "Failed to compute dynamic data: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow"
                ".compute_dynamic_data",
                side_effect=Exception("Boom"),
            ),
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="dynamic-data-failed"
                    )
                ]
            ),
        )

    def test_workflow_wrong_status(self) -> None:
        """A workflow in the wrong status is skipped."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_pending=False
        )
        expected_message = "Failed to mark work request as running"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        # The work request remains blocked.
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)
        # No errors are logged in output_data, because the work request may
        # make progress in the future.
        self.assertIsNone(wr.output_data)

    def test_workflow_fails(self) -> None:
        """A workflow that fails is logged."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        expected_message = "Orchestrator failed: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate",
                side_effect=ValueError("Boom"),
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow_unblocks_children(self) -> None:
        """Workflow children are unblocked if possible."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def populate() -> None:
            children = [wr.create_child_worker("noop") for _ in range(2)]
            children[1].add_dependency(children[0])

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=populate,
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)
        children = wr.children.order_by("id")
        self.assertEqual(children.count(), 2)
        self.assertEqual(children[0].status, WorkRequest.Statuses.PENDING)
        self.assertEqual(children[1].status, WorkRequest.Statuses.BLOCKED)

    def test_wrong_task_type(self) -> None:
        """Attempts to orchestrate non-workflows are logged."""
        wr = self.playground.create_worker_task(
            mark_pending=True, assign_new_worker=True
        )
        expected_message = "Does not have a workflow orchestrator"

        with self.assertLogsContains(
            f"Error running work request {wr.task_type}/{wr.task_name} "
            f"({wr.id}): {expected_message}",
            logger="debusine.server.workflows.base",
            level=logging.WARNING,
        ):
            self.assertFalse(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="configure-failed"
                    )
                ]
            ),
        )


class TaskConfigurationTests(TestCase):
    """test :py:class:`Workflow` task configuration handling."""

    scenario = scenarios.DefaultContext()

    workflow_wr: ClassVar[WorkRequest]

    @classmethod
    def setUpTestData(cls) -> None:
        """Add a sample user to the test fixture."""
        super().setUpTestData()
        cls.workflow_wr = cls.playground.create_workflow()

    def setUp(self) -> None:
        super().setUp()
        workflow = self.workflow_wr.get_task()
        assert isinstance(workflow, Workflow)
        self.workflow = workflow

    def test_ensure_child_worker(self) -> None:
        wr = self.workflow.work_request_ensure_child_worker(
            task_name="noop", workflow_data=WorkRequestWorkflowData()
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_ensure_child_server(self) -> None:
        wr = self.workflow.work_request_ensure_child_server(
            task_name="servernoop", workflow_data=WorkRequestWorkflowData()
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_ensure_child_internal(self) -> None:
        wr = self.workflow.work_request_ensure_child_internal(
            task_name="workflow", workflow_data=WorkRequestWorkflowData()
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_ensure_child_workflow(self) -> None:
        wr = self.workflow.work_request_ensure_child_workflow(
            task_name="noop",
            workflow_data=WorkRequestWorkflowData(),
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_work_request_ensure_child_signing(self) -> None:
        wr = self.workflow.work_request_ensure_child_signing(
            task_name="noop", workflow_data=WorkRequestWorkflowData()
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_work_request_ensure_child_wait(self) -> None:
        package = self.playground.create_minimal_binary_package_artifact()
        wr = self.workflow.work_request_ensure_child_wait(
            task_name="externaldebsign",
            task_data=ExternalDebsignData(unsigned=package.pk),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        self.assertIsNone(wr.configured_task_data)
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)


class RTExampleWorkflowData(BaseWorkflowData):
    """Example Workflow Data for RuntimeParametersTests."""

    dynamic: str
    nested: dict[str, str]
    multiple: list[str]


class RuntimeParametersTests(TestCase):
    """Test Workflow.validate_runtime_parameters()."""

    workflow: ClassVar[
        type[Workflow[RTExampleWorkflowData, BaseDynamicTaskData]]
    ]

    @classmethod
    @preserve_db_task_registry()
    def setUpTestData(cls) -> None:
        """Add a sample Workflow to the test fixture."""
        super().setUpTestData()

        class RTExampleWorkflow(
            SampleWorkflow[RTExampleWorkflowData, BaseDynamicTaskData]
        ):
            TASK_NAME = "example"

        # https://github.com/python/mypy/issues/4717
        cls.workflow = RTExampleWorkflow  # type: ignore[type-abstract]

    def test_validate_any(self) -> None:
        self.workflow.validate_runtime_parameters(RuntimeParameter.ANY)

    def test_validate_empty_dict(self) -> None:
        self.workflow.validate_runtime_parameters({})

    def test_validate_parameter_any(self) -> None:
        self.workflow.validate_runtime_parameters(
            {"dynamic": RuntimeParameter.ANY}
        )

    def test_validate_parameter_values(self) -> None:
        self.workflow.validate_runtime_parameters({"dynamic": ["foo", "bar"]})

    def test_validate_parameter_values_wrong_type(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            r"dynamic is set to \[1\], not a list of valid values\.",
        ):
            self.workflow.validate_runtime_parameters({"dynamic": [1]})

    def test_validate_parameter_single_value(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            (
                r"dynamic is set to a string \('foo'\), not a list of valid "
                r"values\."
            ),
        ):
            self.workflow.validate_runtime_parameters({"dynamic": "foo"})

    def test_validate_unknown_parameter(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            (
                r"unknown_parameter is not a task_data parameter for "
                r"RTExampleWorkflow"
            ),
        ):
            self.workflow.validate_runtime_parameters(
                {"unknown_parameter": RuntimeParameter.ANY}
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""pytest fixtures and hooks for Debusine."""

import os
import shutil
from tempfile import TemporaryDirectory

import pytest

try:
    from pytest_django.lazy_django import django_settings_is_configured
except ImportError:  # integration tests

    def django_settings_is_configured() -> bool:
        """pytest_django has not configured Django."""
        return False


@pytest.hookimpl()
def pytest_configure() -> None:
    """Configure the Prometheus client for multi-process collection."""
    os.environ["PROMETHEUS_MULTIPROC_DIR"] = TemporaryDirectory(
        prefix="debusine-server-prometheus-multiproc-", delete=False
    ).name


@pytest.hookimpl()
def pytest_unconfigure() -> None:  # pragma: no cover
    """Clean up any paths that we created."""
    if django_settings_is_configured():
        from debusine.django.django_utils import cleanup_temp_directory

        cleanup_temp_directory()

    if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
        shutil.rmtree(os.environ["PROMETHEUS_MULTIPROC_DIR"])
        del os.environ["PROMETHEUS_MULTIPROC_DIR"]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Context for database and permission operations.

This is currently based on ContextVar instead of asgiref.local.Local, because
the behaviour of Local is inconsistent across asgiref versions, and probably
buggy in asgiref 3.7+ (see https://github.com/django/asgiref/issues/473).

Ideally, this should eventually aligned with using the same
thread-local/task-local infrastructure that Django uses.

Note that when using threads, if a new thread is started, it will see the
application context reset to empty values: threads that need application
context values need to implement a way to inherit the caller's values.

Note that similar but more complicated gotchas would apply when
`django.urls.reverse` tries to check if `request.urlconf` was previously set to
a non-default value: threads and tasks that need reverse url resolution need to
be mindful of the inconsistencies in behaviour of the underlying
`asgiref.local.Local` implementation, and of asgiref issue #473.

See also https://code.djangoproject.com/ticket/35807
"""

from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import AnonymousUser

if TYPE_CHECKING:
    from debusine.db.models import Scope, Token, User, Workspace
    from debusine.db.models.scopes import ScopeRoles
    from debusine.db.models.workspaces import WorkspaceRoles


class ContextConsistencyError(Exception):
    """Raised if an inconsistency is found when setting application context."""


class Context:
    """
    Storage for Debusine application context.

    Setting the application context needs to be done in a specific order:

    1. ``context.set_scope()``
    2. ``context.set_user()``
    3. ``workspace.set_current()``

    This reflects the order in which information is available in the middleware
    chain, where scope is needed first in order to activate URL resolution,
    which is a prerequisite of the auth layer which will then populate the user.
    """

    # https://github.com/django/asgiref/issues/473

    # Use slots to catch typos in functions that set context variables
    __slots__ = [
        "_permission_checks_disabled",
        "_scope",
        "_scope_roles",
        "_user",
        "_worker_token",
        "_workspace",
        "_workspace_roles",
    ]

    def __init__(self) -> None:
        """Initialize the default values."""
        super().__init__()
        self._permission_checks_disabled: ContextVar[bool] = ContextVar(
            "permission_checks_disabled", default=False
        )

        self._scope: ContextVar[Optional["Scope"]] = ContextVar(
            "scope", default=None
        )

        self._scope_roles: ContextVar[frozenset["ScopeRoles"] | None] = (
            ContextVar("scope_roles", default=None)
        )

        self._user: ContextVar[Union["User", "AnonymousUser", None]] = (
            ContextVar("user", default=None)
        )

        self._worker_token: ContextVar[Optional["Token"]] = ContextVar(
            "worker_token", default=None
        )

        self._workspace: ContextVar[Optional["Workspace"]] = ContextVar(
            "workspace", default=None
        )
        self._workspace_roles: ContextVar[
            frozenset["WorkspaceRoles"] | None
        ] = ContextVar("workspace_roles", default=None)

    @property
    def permission_checks_disabled(self) -> bool:
        """
        Return True if in a management command.

        This is used to skip permission checks in situations like management
        commands or test fixture setup.
        """
        return self._permission_checks_disabled.get()

    @contextmanager
    def disable_permission_checks(self) -> Generator[None]:
        """Set admin mode for the duration of the context manager."""
        orig = self.permission_checks_disabled
        self._permission_checks_disabled.set(True)
        try:
            yield
        finally:
            self._permission_checks_disabled.set(orig)

    @property
    def scope(self) -> Optional["Scope"]:
        """Get the current scope."""
        return self._scope.get()

    def require_scope(self) -> "Scope":
        """
        Get the current scope.

        :raises ContextConsistencyError: if the scope is not set
        """
        if (scope := self.scope) is None:
            raise ContextConsistencyError("scope is not set in context")
        return scope

    def set_scope(self, new_scope: "Scope") -> None:
        """
        Set the current scope.

        This needs to be called *before* ``set_user``.

        :raises ContextConsistencyError: if the scope had already been set
        """
        if (scope := self.scope) is not None:
            raise ContextConsistencyError(f"Scope was already set to {scope}")

        self._scope.set(new_scope)
        self._user.set(None)
        self._workspace.set(None)

    @property
    def user(self) -> Union["User", "AnonymousUser", None]:
        """
        Get the current user.

        :return: the current user, or None if it has not been initialized yet
                 from the request
        """
        return self._user.get()

    def require_user(self) -> Union["User", "AnonymousUser"]:
        """
        Get the current user.

        :raises ContextConsistencyError: if the user is not set
        """
        if (user := self.user) is None:
            raise ContextConsistencyError("user is not set in context")
        return user

    def set_user(self, new_user: Union["User", "AnonymousUser"]) -> None:
        """
        Set the current user.

        This needs to be set *after* ``set_scope``.
        """
        from debusine.db.models.scopes import ScopeRoles

        if (scope := self.scope) is None:
            raise ContextConsistencyError("Cannot set user before scope")

        if (user := self.user) is not None:
            raise ContextConsistencyError(f"User was already set to {user}")

        # TODO: Scope visibility checks should happen here. Currently scopes
        # are always visible, so the only thing needed is this comment as a
        # placeholder

        self._user.set(new_user)
        self._scope_roles.set(
            ScopeRoles.from_iterable(scope.get_group_roles(new_user))
        )

    @property
    def worker_token(self) -> Optional["Token"]:
        """Get the current worker token."""
        return self._worker_token.get()

    def set_worker_token(self, new_token: "Token") -> None:
        """
        Set the current worker token.

        :raises ContextConsistencyError: if the token had already been set
        """
        if not new_token.enabled:
            raise ContextConsistencyError("Token is disabled")
        if self.worker_token is not None:
            raise ContextConsistencyError("Token was already set")
        self._worker_token.set(new_token)

    @property
    def workspace(self) -> Optional["Workspace"]:
        """Get the current workspace."""
        return self._workspace.get()

    def require_workspace(self) -> "Workspace":
        """
        Get the current workspace.

        :raises ContextConsistencyError: if the workspace is not set
        """
        if (workspace := self.workspace) is None:
            raise ContextConsistencyError("workspace is not set in context")
        return workspace

    @property
    def scope_roles(self) -> frozenset["ScopeRoles"]:
        """Get the roles the current user has on the current Scope."""
        if (roles := self._scope_roles.get()) is not None:
            return roles
        raise ContextConsistencyError("user is not set")

    @property
    def workspace_roles(self) -> frozenset["WorkspaceRoles"]:
        """Get the roles the current user has on the current Workspace."""
        if (roles := self._workspace_roles.get()) is not None:
            return roles
        raise ContextConsistencyError("workspace is not set")

    def reset(self) -> None:
        """Reset the application context to default values."""
        self._user.set(None)
        self._worker_token.set(None)
        self._scope.set(None)
        self._scope_roles.set(None)
        self._workspace.set(None)
        self._workspace_roles.set(None)
        self._permission_checks_disabled.set(False)

    @contextmanager
    def local(self) -> Generator[None]:
        """Restore application context when the context manager ends."""
        orig_user = self.user
        orig_worker_token = self.worker_token
        orig_scope = self.scope
        orig_scope_roles = self._scope_roles.get()
        orig_workspace = self.workspace
        orig_workspace_roles = self._workspace_roles.get()
        orig_permission_checks_disabled = self.permission_checks_disabled
        try:
            yield
        finally:
            self._user.set(orig_user)
            self._worker_token.set(orig_worker_token)
            self._scope.set(orig_scope)
            self._scope_roles.set(orig_scope_roles)
            self._workspace.set(orig_workspace)
            self._workspace_roles.set(orig_workspace_roles)
            self._permission_checks_disabled.set(
                orig_permission_checks_disabled
            )


context = Context()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the Worker."""

import asyncio
import concurrent.futures
import datetime as dt
import functools
import logging
import re
import secrets
import signal
import socket
import threading
import time as tm
from collections.abc import AsyncGenerator, Awaitable
from contextlib import asynccontextmanager
from pathlib import Path
from textwrap import dedent
from typing import Any, Literal, override
from unittest import mock
from unittest.mock import MagicMock

import aiohttp
import tenacity
from aiohttp import RequestInfo, WSMessage, WSMsgType
from aiohttp.client_reqrep import ConnectionKey
from aiohttp.web import Response, json_response
from aiohttp.web_exceptions import (
    HTTPForbidden,
    HTTPInternalServerError,
    HTTPServiceUnavailable,
)
from multidict import CIMultiDict, CIMultiDictProxy
from yarl import URL

import debusine.tasks.models as task_models
from debusine.artifacts.models import (
    RuntimeStatistics,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.debusine import Debusine
from debusine.client.exceptions import TokenDisabledError
from debusine.tasks import (
    BaseTask,
    BaseTaskWithExecutor,
    Noop,
    Sbuild,
    TaskConfigError,
)
from debusine.tasks.executors import ExecutorStatistics, UnshareExecutor
from debusine.tasks.models import (
    BackendType,
    OutputData,
    OutputDataError,
    SbuildData,
    SbuildInput,
    WorkerType,
)
from debusine.tasks.tests.helper_mixin import (
    SampleBaseExternalTask,
    SampleBaseTask,
)
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_work_request_response,
)
from debusine.worker import Worker
from debusine.worker.config import ConfigHandler
from debusine.worker.system_information import (
    cpu_time,
    native_architecture,
    system_metadata,
)
from debusine.worker.tests import server
from debusine.worker.tests.server import ServerConfig


class WorkerTests(TestCase, server.DebusineAioHTTPTestCase):
    """Test Worker class."""

    async def setUpAsync(self) -> None:
        """Initialize TestWorker."""
        await super().setUpAsync()
        self.worker: Worker | None = None

        self.api_url = str(self.server.make_url('')) + "/api"

        self.config_temp_directory = self.create_temp_config_directory(
            {'General': {'api-url': self.api_url}}
        )

        self.config = ConfigHandler(
            directories=[self.config_temp_directory.as_posix()]
        )
        self.config['General']['log-file'] = '/dev/null'

        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

    def tearDown(self) -> None:
        """Cleanup after executing a test."""
        super().tearDown()
        # Restore signal handlers. Cli.execute() changes them
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)

    async def tearDownAsync(self) -> None:
        """Asynchronous cleanup."""
        await super().tearDownAsync()
        if self.worker:
            await self.worker.close()

    def setup_valid_token(self) -> None:
        """Set a valid token in the test server and client configuration."""
        token = '6c931875627131b5135b7de3371c44'
        self.server_config.registered_token = token
        self.config.write_token(token)

    def setup_worker(
        self,
        config: ConfigHandler | None = None,
        *,
        connect: bool = False,
        log_file: str | None = None,
        log_level: str | None = None,
        worker_type: Literal[
            WorkerType.EXTERNAL, WorkerType.SIGNING
        ] = WorkerType.EXTERNAL,
        assert_raise_try_again_called: bool = False,
        disable_endless_retry: bool = True,
    ) -> None:
        """Set a new worker to self.worker (with config or self.config)."""
        if config is None:
            config = self.config

        self.worker = Worker(
            log_file=log_file,
            log_level=log_level,
            worker_type=worker_type,
            config=config,
        )
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = Path("/nonexistent")

        try:
            # If the caller of setup_worker created an event loop
            # then let the worker use it.
            self.worker._set_main_event_loop()
        except RuntimeError:
            pass

        if disable_endless_retry:
            raise_try_again_mock = self.disable_endless_retry()

        if connect:
            self.async_run(self.worker.connect())

        if assert_raise_try_again_called:
            raise_try_again_mock.assert_called()

    def patch_http_method(self, http_method: str, **kwargs: Any) -> MagicMock:
        """Patch a specific HTTP method of Worker's HTTP Client."""
        # setup_worker must be called first.
        assert self.worker is not None

        patcher = mock.patch.object(
            self.worker._async_http_client, http_method, autospec=True
        )
        mocked = patcher.start()
        for key, value in kwargs.items():
            setattr(mocked, key, value)
        self.addCleanup(patcher.stop)
        return mocked

    def patch_worker(self, attribute: str, **kwargs: Any) -> mock.AsyncMock:
        """Patch an attribute of the worker object."""
        patcher = mock.patch.object(self.worker, attribute, autospec=True)
        mocked = patcher.start()
        for key, value in kwargs.items():
            setattr(mocked, key, value)
        self.addCleanup(patcher.stop)
        return mocked

    @staticmethod
    def async_run(coroutine: Awaitable[Any]) -> None:
        """Run coroutine in an event loop."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(coroutine)

    def patch_config_handler(self) -> MagicMock:
        """
        Patch ConfigHandler class.

        :return: the MagicMock
        """
        patch = mock.patch(
            "debusine.worker._worker.ConfigHandler", autospec=True
        )
        config_handler_mocked = patch.start()
        self.addCleanup(patch.stop)

        return config_handler_mocked

    def test_worker_init(self) -> None:
        """Init without any keyword."""
        mocked_config_handler = self.patch_config_handler()

        worker = Worker(log_file="/dev/null", log_level="WARNING")

        # Worker() created a ConfigHandler default object and set it
        # to Worker._config
        self.assertIsNotNone(worker._config)
        mocked_config_handler.assert_called_once_with()

    def test_worker_init_signing(self) -> None:
        """Signing workers use different dirs and require an HTTPS api-url."""
        mocked_config_handler = self.patch_config_handler()

        worker = Worker(
            log_file="/dev/null",
            log_level="WARNING",
            worker_type=WorkerType.SIGNING,
        )

        self.assertIsNotNone(worker._config)
        mocked_config_handler.assert_called_once_with(
            directories=[
                str(Path.home() / ".config/debusine/signing"),
                "/etc/debusine/signing",
            ],
            require_https=True,
        )

    def test_worker_init_use_config_handler(self) -> None:
        """worker.set_config_handler() sets the ConfigHandler."""
        self.setup_worker(log_file="/dev/null", log_level="WARNING")
        assert self.worker is not None

        self.assertIs(self.worker._config, self.config)

    def test_registration_success(self) -> None:
        """
        Worker client registers successfully.

        Client creates a token, sends it to the server and save to the
        configuration.
        """
        # No token in the configuration
        self.assertIsNone(self.config.token)

        # Worker connects: it will create a token, send and save
        self.setup_worker(connect=True)

        # A token has been retrieved
        self.assertIsNotNone(self.config.token)

        registration_request = self.server_requests[-2]
        connect_request = self.server_requests[-1]

        self.assertEqual(registration_request.path, '/api/1.0/worker/register/')
        self.assertEqual(connect_request.path, '/api/ws/1.0/worker/connect/')

        # Server side received expected registration data
        self.assertEqual(
            registration_request.json_content,
            {
                'token': self.config.token,
                'fqdn': socket.getfqdn(),
                'worker_type': WorkerType.EXTERNAL,
            },
        )
        assert registration_request.headers is not None
        self.assertNotIn('token', registration_request.headers)

    def test_registration_success_activation_token(self) -> None:
        """Worker client registers successfully with an activation token."""
        activation_token = secrets.token_hex(32)
        activation_token_path = Path(self.config._activation_token_file)
        activation_token_path.write_bytes(activation_token.encode())
        activation_token_path.chmod(0o600)
        self.assertEqual(self.config.activation_token, activation_token)

        # No token in the configuration
        self.assertIsNone(self.config.token)

        # Worker connects: it will create a token, send and save
        self.setup_worker(connect=True)

        # A token has been retrieved
        self.assertIsNotNone(self.config.token)

        registration_request = self.server_requests[-2]
        connect_request = self.server_requests[-1]

        self.assertEqual(registration_request.path, '/api/1.0/worker/register/')
        self.assertEqual(connect_request.path, '/api/ws/1.0/worker/connect/')

        # Server side received expected registration data
        self.assertEqual(
            registration_request.json_content,
            {
                'token': self.config.token,
                'fqdn': socket.getfqdn(),
                'worker_type': WorkerType.EXTERNAL,
            },
        )
        assert registration_request.headers is not None
        self.assertEqual(
            registration_request.headers['token'], activation_token
        )

    def test_registration_failure(self) -> None:
        """
        Worker client registration failure.

        The server returns 403, the client raises WorkerRegistrationError.
        """
        self.server_config.register_response_status_code = (
            HTTPForbidden.status_code
        )

        log_message = (
            'Could not register. Server HTTP status code: 403 Body: error'
        )

        self.setup_worker()
        assert self.worker is not None
        with (
            self.assertLogsContains(log_message),
            self.assertRaisesSystemExit(3),
        ):
            self.async_run(self.worker.connect())

    def test_connection_success(self) -> None:
        """Worker client connects successfully."""
        self.setup_valid_token()

        self.setup_worker(connect=True)
        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/ws/1.0/worker/connect/'
        )
        self.assertTrue(self.server_latest_request.is_authenticated)

    def disable_endless_retry(self) -> mock.MagicMock:
        """Disable Worker._try_again() raising tenacity.RetryAgain."""
        patcher = mock.patch.object(
            self.worker, "_raise_try_again", autospec=True
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)

        return mocked

    def test_connection_failure(self) -> None:
        """Worker client cannot connect: invalid token."""
        self.registered_token = 'f924ba0099588bf6'

        self.config.write_token('e3c0284')

        self.setup_worker(connect=True, assert_raise_try_again_called=True)

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/ws/1.0/worker/connect/'
        )
        self.assertFalse(self.server_latest_request.is_authenticated)

    async def test_connection_reset_error(self) -> None:
        """Worker client handles ConnectionResetError exception."""
        self.setup_valid_token()
        self.setup_worker()
        assert self.worker is not None

        self.patch_worker("_process_message", side_effect=ConnectionResetError)
        mock_wait_before_reconnect = self.patch_worker(
            "_asyncio_sleep",
            side_effect=[None],
        )

        self.server_config.connect_response_bodies = [{}]

        log_message = 'ConnectionResetError, will reconnect'

        # It tried twice to connect: once before the call to
        # Worker._wait_before_reconnect and once when it has been executed
        with (
            self.assertLogsContains(
                log_message, expected_count=2, level=logging.DEBUG
            ),
            self.assertRaises(StopAsyncIteration),
        ):
            await self.worker.connect()

        # It's called twice: first it returns None and the second time raises
        # and exception because side_effect=[None]
        self.assertEqual(mock_wait_before_reconnect.call_count, 2)

    async def test_connection_read_error(self) -> None:
        """Worker cannot read data from the server (low level disconnection)."""
        self.setup_worker()
        assert self.worker is not None
        mock_post = self.patch_http_method(
            "post",
            side_effect=aiohttp.client_exceptions.ClientResponseError(
                RequestInfo(
                    URL('https://test.com'),
                    '',
                    CIMultiDictProxy(CIMultiDict()),
                    URL('https://test.com'),
                ),
                (),
            ),
        )
        log_message = (
            'Could not register. Server HTTP status code: 0 Body: Not available'
        )
        with (
            self.assertLogsContains(log_message),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.connect()

        self.assertTrue(mock_post.called)

    async def test_client_connector_error(self) -> None:
        """Worker cannot connect to the server: server unreachable."""
        self.setup_worker()
        assert self.worker is not None
        connection_key = mock.create_autospec(spec=ConnectionKey)
        mock_post = self.patch_http_method(
            "post",
            side_effect=aiohttp.client_exceptions.ClientConnectorError(
                connection_key, OSError("Unreachable")
            ),
        )
        log_message = "Could not register. Server unreachable: "
        with (
            self.assertLogsContains(log_message),
            self.assertRaisesSystemExit(3),
        ):
            await self.worker.connect()

        self.assertTrue(mock_post.called)

    def test_connection_rejected(self) -> None:
        """Worker tries to connect, server returns 500."""
        self.server_config.register_response_status_code = (
            HTTPInternalServerError.status_code
        )

        log_message = (
            'Could not register. Server HTTP status code: 500 Body: error'
        )

        self.setup_worker()
        assert self.worker is not None

        with (
            self.assertLogsContains(log_message),
            self.assertRaisesSystemExit(3),
        ):
            self.async_run(self.worker.connect())

    async def test_connection_error_connection_refused(self) -> None:
        """Worker client cannot connect: connection refused."""
        # Because the token exists the worker will try to connect
        # instead of to register
        self.config.write_token('e3c0284')

        await self.server.close()

        self.setup_worker()
        assert self.worker is not None
        mocked_sleep = self.patch_worker(
            "_asyncio_sleep", side_effect=[None, None]
        )

        log_message = 'Error connecting to'

        # Assert that it tried three times to connect
        # The first one Work._wait_before_reconnect wasn't called, then
        # twice before Work._wait_before_reconnect returns None twice

        with (
            self.assertLogsContains(log_message, expected_count=3),
            self.assertRaises(StopAsyncIteration),
        ):
            # StopAsyncIteration because mocked_sleep side_effects
            await self.worker.connect()

        # It's called three times: twice return None and once raises
        # and exception because mocked_sleep.side_effect=[None, None]
        self.assertEqual(mocked_sleep.call_count, 3)

    async def test_do_wait_for_messages_raise_value_error(self) -> None:
        """Exception ValueError is raised (not retried via tenacity)."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['token_disabled_error']
        ]

        self.setup_worker()
        assert self.worker is not None

        exception = ValueError
        self.patch_worker("_process_message", side_effect=exception)

        with self.assertRaises(exception):
            await self.worker.connect()

    async def test_close_exits_connect_loop(self) -> None:
        """The main loop in connect() stops if we explicitly close."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['token_disabled_error']
        ]

        self.setup_worker(disable_endless_retry=False)
        assert self.worker is not None

        async def process_message(
            msg: aiohttp.WSMessage,  # noqa: ARG001, U100
            client: Debusine,  # noqa: ARG001, U100
        ) -> bool:
            assert self.worker is not None
            await self.worker.close()
            return False

        with (
            mock.patch.object(
                self.worker, "_process_message", side_effect=process_message
            ),
            mock.patch.object(
                self.worker, "_raise_try_again"
            ) as try_again_mock,
        ):
            await self.worker.connect()

        try_again_mock.assert_not_called()

    async def test_connection_error_token_disabled_try_again(self) -> None:
        """Worker tries to connect again if server returns "TOKEN_DISABLED"."""
        self.setup_valid_token()
        assert self.config.token is not None
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['token_disabled_error']
        ]

        # No request to any URL...
        self.assertIsNone(self.server_latest_request)

        self.setup_worker()
        assert self.worker is not None
        mocked_method = self.patch_worker(
            '_asyncio_sleep', side_effect=[None, None]
        )

        with (
            self.assertRaises(StopAsyncIteration),
            self.assertLogsContains(
                self.config.token, expected_count=3
            ) as logs,
        ):
            # RuntimeError due to the implementation of the mocked
            # returning None twice
            await self.worker.connect()

        # It's called three times: twice _wait_before_reconnected returned None,
        # and once raises StopIteration
        self.assertEqual(mocked_method.call_count, 3)

        # Assert that the logs contain "Reason: The token is disabled"
        token_is_disabled_found = False
        for log in logs.output:  # pragma: no cover
            if "Reason: The token is disabled" in log:
                token_is_disabled_found = True
                break

        self.assertTrue(token_is_disabled_found)

    async def test_process_message_dynamic_metadata_request(self) -> None:
        """Worker receives a request to send the dynamic metadata."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['request_dynamic_metadata']
        ]

        # No request to any URL...
        self.assertIsNone(self.server_latest_request)

        self.setup_worker()
        assert self.worker is not None
        send_dynamic_metadata = self.patch_worker("_send_dynamic_metadata")

        await self.worker.connect()

        # The server sends a 'request_dynamic_metadata' and the worker
        # sends system_metadata(task_models.WorkerType.EXTERNAL),
        # analyze_worker_all_executors, and analyze_external_worker_tasks
        send_dynamic_metadata.assert_called_once()

    async def test_process_message_dynamic_metadata_request_signing(
        self,
    ) -> None:
        """Signing workers send the correct worker type in dynamic metadata."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES["request_dynamic_metadata"]
        ]
        self.assertIsNone(self.server_latest_request)

        self.setup_worker(worker_type=WorkerType.SIGNING)
        assert self.worker is not None
        with (
            mock.patch(
                "debusine.worker._worker.analyze_worker_all_executors",
                return_value={},
            ),
            mock.patch(
                "debusine.worker._worker.analyze_external_worker_tasks",
                return_value={},
            ),
            mock.patch(
                "debusine.worker._worker.Worker._get_provided_tags",
                autospec=True,
                return_value=set(),
            ),
        ):
            await self.worker.connect()

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, "/api/1.0/worker/dynamic-metadata/"
        )
        self.assertTrue(self.server_latest_request.is_authenticated)
        self.assertEqual(
            self.server_latest_request.json_content,
            {
                **system_metadata(task_models.WorkerType.SIGNING),
                "provided_tags": [],
            },
        )

    async def test_send_dynamic_metadata(self) -> None:
        """Test a successful dynamic metadata update."""
        worker_metadata = {"test": True}

        self.setup_worker()
        self.setup_valid_token()
        assert self.worker is not None

        with (
            mock.patch(
                "debusine.worker._worker.Worker._get_worker_metadata",
                autospec=True,
                return_value=worker_metadata,
            ),
            self.assertLogs(level=logging.DEBUG) as logs,
        ):
            await self.worker._send_dynamic_metadata({})

        regex = r'Sent dynamic_metadata \(response: 204\) Dynamic metadata: .*'
        found = sum(
            int(re.search(regex, log) is not None) for log in logs.output
        )
        self.assertEqual(found, 1)

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/1.0/worker/dynamic-metadata/'
        )

        self.assertTrue(self.server_latest_request.is_authenticated)

        self.assertEqual(
            self.server_latest_request.json_content, worker_metadata
        )

    async def test_send_dynamic_metadata_failed(self) -> None:
        """Test a failed dynamic metadata update."""
        executors_metadata = {"executor:unshare:available": True}
        tasks_metadata = {"sbuild:version": 1, "sbuild:available": True}

        self.setup_worker()
        self.setup_valid_token()
        assert self.worker is not None

        self.server_config.dynamic_metadata_status_code = (
            HTTPServiceUnavailable.status_code
        )

        with (
            mock.patch(
                "debusine.worker._worker.analyze_worker_all_executors",
                autospec=True,
                return_value=executors_metadata,
            ),
            mock.patch(
                'debusine.worker._worker.analyze_external_worker_tasks',
                autospec=True,
                return_value=tasks_metadata,
            ),
            self.assertLogsContains(
                'INFO:root:Failed to send dynamic_metadata: (response: 503): '
            ),
        ):
            await self.worker._send_dynamic_metadata({})

        assert self.server_latest_request is not None
        self.assertEqual(
            self.server_latest_request.path, '/api/1.0/worker/dynamic-metadata/'
        )

    def assert_request_done_times(
        self, method: str, path: str, times: int
    ) -> None:
        """Assert that request method regexp_path has been done times."""
        count = 0
        for request in self.server_requests:
            if request.method == method and request.path == path:
                count += 1

        self.assertEqual(count, times)

    async def test_poll_statistics(self) -> None:
        """`_poll_statistics` accumulates statistics."""
        self.setup_worker()
        assert self.worker is not None
        self.worker._task_running = Sbuild(
            SbuildData(
                input=SbuildInput(source_artifact=1),
                build_architecture="amd64",
                backend=BackendType.UNSHARE,
                environment="debian/match:codename=sid",
            ).model_dump()
        )
        with mock.patch(
            "debusine.tasks.executors.images.ImageCache.image_artifact",
            return_value=create_artifact_response(id=1),
        ):
            self.worker._task_running.executor = UnshareExecutor(
                mock.create_autospec(Debusine), 1, workspace="workspace"
            )
        self.worker._task_statistics = RuntimeStatistics()
        GiB = 1024 * 1024 * 1024
        got_statistics = [
            ExecutorStatistics(),
            ExecutorStatistics(
                disk_space=GiB,
                memory=GiB,
                available_disk_space=10 * GiB,
                available_memory=4 * GiB,
                cpu_count=2,
            ),
            ExecutorStatistics(
                disk_space=3 * GiB,
                memory=2 * GiB,
                available_disk_space=10 * GiB,
                available_memory=4 * GiB,
                cpu_count=2,
            ),
            ExecutorStatistics(
                disk_space=4 * GiB,
                memory=3 * GiB,
                available_disk_space=10 * GiB,
                available_memory=4 * GiB,
                cpu_count=2,
            ),
            ExecutorStatistics(
                disk_space=2 * GiB,
                memory=GiB,
                available_disk_space=10 * GiB,
                available_memory=4 * GiB,
                cpu_count=2,
            ),
        ]
        sleep_results = [None] * (len(got_statistics) - 1) + [
            asyncio.CancelledError
        ]

        with (
            self.assertRaises(asyncio.CancelledError),
            mock.patch(
                "debusine.tasks.executors.unshare.UnshareExecutor"
                ".get_statistics",
                side_effect=got_statistics,
            ),
            mock.patch("asyncio.sleep", side_effect=sleep_results),
        ):
            await self.worker._poll_statistics()

        self.assertEqual(
            self.worker._task_statistics,
            RuntimeStatistics(
                disk_space=4 * GiB,
                memory=3 * GiB,
                available_disk_space=10 * GiB,
                available_memory=4 * GiB,
                cpu_count=2,
            ),
        )

    async def test_poll_statistics_no_executor(self) -> None:
        """`_poll_statistics` ignores tasks without an executor."""
        self.setup_worker()
        assert self.worker is not None
        self.worker._task_running = Noop({"result": WorkRequestResults.SUCCESS})
        self.worker._task_statistics = RuntimeStatistics()

        with mock.patch("asyncio.sleep") as mock_sleep:
            await self.worker._poll_statistics()

        mock_sleep.assert_not_called()
        self.assertEqual(self.worker._task_statistics, RuntimeStatistics())

    @asynccontextmanager
    async def completed_task(
        self, *, keep_state: bool = False
    ) -> AsyncGenerator[None]:
        """
        Context manager that waits for a task to complete.

        :param keep_state: If True, don't reset state at the end of the
          task, to allow tests to inspect it.
        """
        assert self.worker is not None
        assert self.worker._main_event_loop is not None

        completed = asyncio.Event()
        original_reset_state = self.worker._reset_state

        async def report_completed(worker: Worker) -> None:
            if keep_state:
                # Worker._reset_state would do this; we should clean up too.
                if worker._task_lock.locked():  # pragma: no cover
                    worker._task_lock.release()
            else:
                original_reset_state()

            completed.set()

        with mock.patch(
            "debusine.worker._worker.Worker._reset_state",
            side_effect=functools.partial(
                asyncio.run_coroutine_threadsafe,
                report_completed(self.worker),
                self.worker._main_event_loop,
            ),
        ):
            async with asyncio.timeout(10):
                waiter_task = asyncio.create_task(completed.wait())
                yield
                await waiter_task

    async def receive_task_submit_result(
        self,
        task_type: TaskTypes,
        task_name: str,
        execute_result: WorkRequestResults,
        json_result: str,
        worker_type: Literal[
            WorkerType.EXTERNAL, WorkerType.SIGNING
        ] = WorkerType.EXTERNAL,
    ) -> None:
        """Worker fetches the task and validates submitted JSON."""
        task_class = BaseTask.class_from_name(task_type, task_name)

        patcher_execute = mock.patch.object(
            task_class, 'execute', autospec=True
        )
        mocked_execute = patcher_execute.start()
        mocked_execute.return_value = execute_result
        self.addCleanup(patcher_execute.stop)

        patcher_configure_server_access = mock.patch.object(
            task_class, "configure_server_access", autospec=True
        )
        mocked_configure_server_access = patcher_configure_server_access.start()
        self.addCleanup(patcher_configure_server_access.stop)

        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['work_request_available']
        ]

        # No request to any URL...
        self.assertIsNone(self.server_latest_request)

        self.setup_worker(worker_type=worker_type)
        assert self.worker is not None

        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )

        with mock.patch(
            "debusine.tasks.executors.images.ImageCache.image_artifact",
            return_value=create_artifact_response(id=1),
        ):
            executor = UnshareExecutor(
                mock.create_autospec(Debusine), 1, workspace="workspace"
            )
        GiB = 1024 * 1024 * 1024

        with (
            mock.patch(
                "debusine.tasks._task.BaseTaskWithExecutor.executor",
                create=True,
                new_callable=mock.PropertyMock,
                return_value=executor,
            ),
            mock.patch.object(
                executor,
                "get_statistics",
                return_value=ExecutorStatistics(
                    disk_space=2 * GiB,
                    memory=GiB,
                    available_disk_space=20 * GiB,
                    available_memory=8 * GiB,
                    cpu_count=4,
                ),
            ),
        ):
            async with self.completed_task(keep_state=True):
                await self.worker.connect()

        # The WorkRequest is executed...
        self.assertEqual(mocked_execute.call_count, 1)

        # Check the workspace and worker native architecture assigned to the
        # task
        task = mocked_execute.call_args[0][0]
        self.assertEqual(task.workspace_name, "TestWorkspace")
        self.assertEqual(task.worker_native_architecture, native_architecture())

        # BaseTask.configure_server_access was called
        debusine_server = mocked_configure_server_access.mock_calls[0][1][1]
        self.assertIsInstance(debusine_server, Debusine)
        self.assertEqual(debusine_server.token, self.config.token)
        self.assertEqual(debusine_server.base_api_url, self.config.api_url)

        work_request_completed_mocked.assert_called_once()
        call_args = work_request_completed_mocked.call_args[0]

        client = call_args[0]  # the self arg
        self.assertEqual(client.scope, "testscope")
        self.assertEqual(call_args[1], task.work_request_id)
        self.assertEqual(call_args[2], json_result)

        output_data = call_args[3]
        self.assertIsInstance(output_data, OutputData)
        runtime_statistics = output_data.runtime_statistics
        assert runtime_statistics is not None
        assert self.worker._task_start_time is not None
        assert self.worker._task_start_cpu_time is not None
        self.assertLessEqual(
            runtime_statistics.duration,
            tm.monotonic() - self.worker._task_start_time,
        )
        self.assertLessEqual(
            runtime_statistics.cpu_time,
            cpu_time() - self.worker._task_start_cpu_time,
        )
        if isinstance(task, BaseTaskWithExecutor):
            self.assertEqual(runtime_statistics.disk_space, 2 * GiB)
            self.assertEqual(runtime_statistics.memory, GiB)
            self.assertEqual(runtime_statistics.available_disk_space, 20 * GiB)
            self.assertEqual(runtime_statistics.available_memory, 8 * GiB)
            self.assertEqual(runtime_statistics.cpu_count, 4)
        else:
            self.assertIsNone(runtime_statistics.disk_space)
            self.assertIsNone(runtime_statistics.memory)
            self.assertIsNone(runtime_statistics.available_disk_space)
            self.assertIsNone(runtime_statistics.available_memory)
            self.assertIsNone(runtime_statistics.cpu_count)

    async def test_task_is_received_executed_submit_success(self) -> None:
        """Receive a task, execute it and submit success completion."""
        await self.receive_task_submit_result(
            TaskTypes.WORKER, "sbuild", WorkRequestResults.SUCCESS, "success"
        )

    async def test_task_is_received_executed_submit_failure(self) -> None:
        """Receive a task, executes it (failure) and submits failure."""
        await self.receive_task_submit_result(
            TaskTypes.WORKER, "sbuild", WorkRequestResults.FAILURE, "failure"
        )

    async def test_print_no_work_request_available(self) -> None:
        """Worker logs 'No work request available'."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['work_request_available']
        ]
        self.server_config.get_next_for_worker = (
            ServerConfig.GetNextForWorkerResponse.NO_WORK_REQUEST_PENDING
        )

        self.setup_worker()
        assert self.worker is not None

        log_message = 'No work request available'

        with self.assertLogsContains(log_message, level=logging.DEBUG):
            await self.worker.connect()

    async def test_log_work_request_ignored_already_running(self) -> None:
        """Worker receive two Work Request available, fetches only one."""
        lock = threading.Lock()
        lock.acquire()

        def wait(sbuild_self: Sbuild) -> None:
            sbuild_self  # fake usage for vulture
            lock.acquire()
            lock.release()

        patcher = mock.patch.object(Sbuild, "execute", autospec=True)
        mocked_execute = patcher.start()
        mocked_execute.side_effect = wait
        self.addCleanup(patcher.stop)

        # avoid NotFoundError in test env
        patcher = mock.patch(
            "debusine.client.debusine.Debusine.work_request_completed_update"
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        self.setup_valid_token()

        work_request_available = ServerConfig.RESPONSES[
            "work_request_available"
        ]

        self.server_config.connect_response_bodies = [
            work_request_available,
            work_request_available,
        ]

        self.setup_worker()
        assert self.worker is not None
        self.patch_worker("_asyncio_sleep")
        mocked_lock = self.patch_worker("_task_lock")
        mocked_lock.acquire.side_effect = [True] + [False] * 61

        log_message = "Worker is busy and can't execute a new work request"

        async with self.completed_task():
            # Log that a Work Request is available (coming from the server)
            # but the Worker ignores it because the worker is already
            # executing a task
            with self.assertLogsContains(log_message):
                await self.worker.connect()

            # Sbuild.execute() is called only once
            self.assertEqual(mocked_execute.call_count, 1)

            # The server sent twice "work_request_available" and the worker
            # called get-next-for-worker/ once only: because a Task is
            # already being executed
            self.assert_request_done_times(
                "GET", "/api/1.0/work-request/get-next-for-worker/", times=1
            )

            # Allow the execute call to finish
            lock.release()

    async def test_log_work_request_cannot_report_exits(self) -> None:
        """Worker can't report to the server, so logs this and exits."""
        # Mock Sbuild.execute()
        patcher_execute = mock.patch.object(Sbuild, 'execute', autospec=True)
        mocked_execute = patcher_execute.start()
        mocked_execute.return_value = True
        self.addCleanup(patcher_execute.stop)

        patcher_configure_server_access = mock.patch.object(
            Sbuild, "configure_server_access", autospec=True
        )
        patcher_configure_server_access.start()
        self.addCleanup(patcher_configure_server_access.stop)

        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['work_request_available']
        ]

        self.setup_worker()
        assert self.worker is not None

        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )
        work_request_completed_mocked.side_effect = Exception("network error")

        with (
            self.assertLogsContains("Cannot reach server", level=logging.ERROR),
            mock.patch(
                "debusine.worker.Worker._exit_worker", autospec=True
            ) as exit_worker,
        ):
            async with self.completed_task():
                await self.worker.connect()
            exit_worker.assert_called_once()

    async def assert_invalid_get_next_for_worker_logging(
        self, get_next_for_worker: ServerConfig.GetNextForWorkerResponse
    ) -> None:
        """Assert log_message is logged for get_next_for_worker response."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['work_request_available']
        ]
        self.server_config.get_next_for_worker = get_next_for_worker

        self.setup_worker()
        assert self.worker is not None

        with self.assertLogsContains(
            "Invalid WorkRequest received from", level=logging.WARNING
        ):
            await self.worker.connect()

    async def test_request_work_request_invalid_work_request_response(
        self,
    ) -> None:
        """Worker logs 'Invalid content of...'."""
        await self.assert_invalid_get_next_for_worker_logging(
            ServerConfig.GetNextForWorkerResponse.INVALID_WORK_REQUEST
        )

    async def test_request_work_request_invalid_json_response(self) -> None:
        """Worker logs 'Invalid JSON response...'."""
        await self.assert_invalid_get_next_for_worker_logging(
            ServerConfig.GetNextForWorkerResponse.INVALID_JSON
        )

    async def test_request_work_request_retries(self) -> None:
        """_request_work_request retires failed requests."""
        self.server_config.get_next_for_worker = Response(
            status=HTTPServiceUnavailable.status_code
        )
        self.setup_worker()
        assert self.worker is not None

        def sleep(delay: float) -> None:  # noqa: ARG001, U100
            self.server_config.get_next_for_worker = (
                ServerConfig.GetNextForWorkerResponse.VALID_WORK_REQUEST
            )

        self.patch_worker("_asyncio_sleep", side_effect=sleep)
        response = await self.worker._request_work_request()
        self.assertIsNotNone(response)

    async def assert_execute_work_request_send_and_log_error(
        self,
        error_msg: str,
        error_code: str,
        worker_type: Literal[
            WorkerType.EXTERNAL, WorkerType.SIGNING
        ] = WorkerType.EXTERNAL,
    ) -> None:
        """Assert error is sent to the server and logged."""
        self.setup_valid_token()
        self.server_config.connect_response_bodies = [
            ServerConfig.RESPONSES['work_request_available']
        ]

        self.setup_worker(worker_type=worker_type)
        assert self.worker is not None

        work_request_completed_mocked = (
            self.patch_debusine_work_request_completed_update()
        )

        with self.assertLogsContains(error_msg):
            async with self.completed_task():
                await self.worker.connect()

        work_request_completed_mocked.assert_called_once_with(
            mock.ANY, mock.ANY, "error", mock.ANY
        )
        _, work_request_id, _, output_data = (
            work_request_completed_mocked.call_args[0]
        )
        self.assertIsInstance(work_request_id, int)
        self.assertIsInstance(output_data, OutputData)
        self.assertEqual(
            output_data.errors,
            [OutputDataError(message=error_msg, code=error_code)],
        )

    async def test_task_configure_error_submit_error(self) -> None:
        """Sbuild.configure fails, Worker submits error."""
        patcher = mock.patch.object(Sbuild, '_configure', autospec=True)
        mocked_configure = patcher.start()
        mocked_configure.side_effect = TaskConfigError('Invalid schema')
        self.addCleanup(patcher.stop)

        await self.assert_execute_work_request_send_and_log_error(
            "Task: sbuild Error configure: Invalid schema", "configure-failed"
        )

    async def test_server_task_on_external_worker(self) -> None:
        """An external worker refuses to run a server task."""

        class ServerTestData(task_models.BaseTaskData):
            pass

        class ServerTest(
            SampleBaseTask[ServerTestData, task_models.BaseDynamicTaskData],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SERVER

            @override
            def _execute(self) -> WorkRequestResults:
                raise NotImplementedError()

            @override
            def _upload_work_request_debug_logs(self) -> None:
                raise NotImplementedError()

        self.addCleanup(
            BaseTask._sub_tasks[TaskTypes.SERVER].__delitem__, "servertest"
        )

        next_work_request = create_work_request_response(
            id=52,
            task_type="Server",
            task_name="servertest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.server_config.get_next_for_worker = json_response(
            next_work_request.model_dump(mode="json")
        )

        await self.assert_execute_work_request_send_and_log_error(
            "Task: servertest is of type Server, not Worker", "wrong-task-type"
        )

    async def test_signing_task_on_external_worker(self) -> None:
        """An external worker refuses to run a signing task."""

        class SigningTestData(task_models.BaseTaskData):
            pass

        class SigningTest(
            SampleBaseExternalTask[
                SigningTestData, task_models.BaseDynamicTaskData
            ],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SIGNING

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
                raise NotImplementedError()

        self.addCleanup(
            BaseTask._sub_tasks[TaskTypes.SIGNING].__delitem__, "signingtest"
        )

        next_work_request = create_work_request_response(
            id=52,
            task_type="Signing",
            task_name="signingtest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.server_config.get_next_for_worker = json_response(
            next_work_request.model_dump(mode="json")
        )

        await self.assert_execute_work_request_send_and_log_error(
            "Task: signingtest is of type Signing, not Worker",
            "wrong-task-type",
        )

    async def test_worker_task_on_signing_worker(self) -> None:
        """A signing worker refuses to run a worker task."""
        next_work_request = create_work_request_response(
            id=52,
            task_type="Worker",
            task_name="noop",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.server_config.get_next_for_worker = json_response(
            next_work_request.model_dump(mode="json")
        )

        await self.assert_execute_work_request_send_and_log_error(
            "Task: noop is of type Worker, not Signing",
            "wrong-task-type",
            worker_type=WorkerType.SIGNING,
        )

    async def test_signing_task_on_signing_worker(self) -> None:
        """A signing worker can run a signing task."""

        class SigningTestData(task_models.BaseTaskData):
            pass

        class SigningTest(
            SampleBaseExternalTask[
                SigningTestData, task_models.BaseDynamicTaskData
            ],
        ):
            TASK_VERSION = 1
            TASK_TYPE = TaskTypes.SIGNING

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
                raise NotImplementedError()

        self.addCleanup(
            BaseTask._sub_tasks[TaskTypes.SIGNING].__delitem__, "signingtest"
        )

        next_work_request = create_work_request_response(
            id=52,
            task_type="Signing",
            task_name="signingtest",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.server_config.get_next_for_worker = json_response(
            next_work_request.model_dump(mode="json")
        )

        await self.receive_task_submit_result(
            TaskTypes.SIGNING,
            "signingtest",
            WorkRequestResults.SUCCESS,
            "success",
            worker_type=WorkerType.SIGNING,
        )

    async def test_unknown_task_name(self) -> None:
        """A worker refuses to run a task with an unknown name."""
        next_work_request = create_work_request_response(
            id=52,
            task_type="Worker",
            task_name="nonexistent",
            created_at=dt.datetime.fromisoformat("2022-01-05T11:14:22.242178Z"),
            started_at=None,
            completed_at=None,
            duration=None,
            worker=58,
            task_data={},
            dynamic_task_data={},
            priority_base=0,
            priority_adjustment=0,
            status="running",
            result="",
            artifacts=[],
            workspace="TestWorkspace",
        )
        self.server_config.get_next_for_worker = json_response(
            next_work_request.model_dump(mode="json")
        )

        await self.assert_execute_work_request_send_and_log_error(
            "Task: nonexistent Error setup:"
            " 'nonexistent' is not a registered Worker task_name",
            "setup-failed",
        )

    def patch_debusine_work_request_completed_update(self) -> MagicMock:
        """Patch Debusine.work_request_completed_update. Return its mock."""
        # setup_worker must be called first.
        assert self.worker is not None

        patcher = mock.patch(
            "debusine.client.debusine.Debusine.work_request_completed_update",
            autospec=True,
        )
        mocked = patcher.start()
        self.addCleanup(patcher.stop)
        return mocked

    async def test_task_execution_error_submit_error(self) -> None:
        """Sbuild.execute raises TaskConfigError, Worker submits error."""
        patcher_execute = mock.patch.object(Sbuild, 'execute', autospec=True)
        mocked_execute = patcher_execute.start()
        mocked_execute.side_effect = TaskConfigError('Broken')
        self.addCleanup(patcher_execute.stop)

        await self.assert_execute_work_request_send_and_log_error(
            'Task: sbuild Error execute: Broken', "execute-failed"
        )

    async def test_process_message_invalid_msg(self) -> None:
        """Worker._process_message returns False."""
        msg = WSMessage(WSMsgType.TEXT, '{"text": "something"}', None)

        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        message_processed = await self.worker._process_message(msg, client)

        self.assertFalse(message_processed)

    async def test_process_message_txt_without_msg_key(self) -> None:
        """Worker._process_message returns False (no msg text)."""
        msg = WSMessage(WSMsgType.TEXT, '{"foo": "bar"}', None)
        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        log_message = "Disconnected. Reason: 'unknown'"

        with self.assertLogsContains(log_message):
            message_processed = await self.worker._process_message(msg, client)

        self.assertFalse(message_processed)

    async def test_process_message_raise_token_disabled_error_disabled(
        self,
    ) -> None:
        """Worker._process_message logs reason of worker's not connection."""
        msg = WSMessage(
            WSMsgType.TEXT,
            '{"reason": "Disabled token", "reason_code": "TOKEN_DISABLED"}',
            None,
        )
        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        with self.assertRaises(TokenDisabledError) as exc:
            await self.worker._process_message(msg, client)

        self.assertEqual(str(exc.exception), "Disabled token")

    async def test_process_message_raise_token_disabled_error_unknown(
        self,
    ) -> None:
        """Worker._process_message logs reason of worker's, exc is 'Unknown'."""
        msg = WSMessage(
            WSMsgType.TEXT,
            '{"reason_code": "TOKEN_DISABLED"}',
            None,
        )
        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        with self.assertRaises(TokenDisabledError) as exc:
            await self.worker._process_message(msg, client)

        self.assertEqual(str(exc.exception), "Unknown")

    async def test_process_message_unknown_reason_code_is_logged_ignored(
        self,
    ) -> None:
        """Worker logs and returns False for msg with an unknown reason_code."""
        msg = WSMessage(
            WSMsgType.TEXT,
            '{"reason": "Disabled token",'
            '"reason_code": "SOMETHING_NOT_IMPLEMENTED"}',
            None,
        )

        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        with self.assertLogsContains('SOMETHING_NOT_IMPLEMENTED'):
            processed = await self.worker._process_message(msg, client)

        self.assertFalse(processed)

    async def test_process_message_non_text_type(self) -> None:
        """Worker._process_message returns False for non-text type."""
        msg = WSMessage(WSMsgType.PING, None, None)
        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        message_processed = await self.worker._process_message(msg, client)

        self.assertFalse(message_processed)

    async def test_process_message_invalid_msg_type(self) -> None:
        """Worker._process_message return False for an invalid message type."""
        msg = 'Something not an instance of WMessage'

        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        log_message = (
            "Worker._process_message: unexpected type: <class 'str'> is not an "
            "instance of "
        )
        with self.assertLogsContains(log_message, level=logging.DEBUG):
            message_processed = await self.worker._process_message(
                msg,  # type: ignore[arg-type]
                client,
            )

        self.assertFalse(message_processed)

    async def test_process_message_invalid_json(self) -> None:
        """Worker._process_message return False if the JSON is invalid."""
        msg = WSMessage(WSMsgType.TEXT, '{[', None)

        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        message_processed = await self.worker._process_message(msg, client)

        self.assertFalse(message_processed)

    def patch_config_handler_fail(
        self, config_handler: ConfigHandler
    ) -> MagicMock:
        """
        Patch ConfigHandler._fail and returns mock.

        side_effect is aligned with the behaviour of _fail.
        """
        patched = mock.patch.object(config_handler, '_fail', autospec=True)
        fail_mocked = patched.start()
        fail_mocked.side_effect = SystemExit(3)
        self.addCleanup(patched.stop)

        return fail_mocked

    async def test_invalid_configuration_missing_debusine_url(self) -> None:
        """Worker() raises an exception: api_url is not in [General]."""
        config_temp_directory = self.create_temp_config_directory(
            {'General': {}}
        )
        config = ConfigHandler(directories=[config_temp_directory.as_posix()])

        log_message = (
            f'Missing required key "api-url" in General '
            f'section in the configuration file '
            f'({config.active_configuration_file})'
        )

        fail_mocked = self.patch_config_handler_fail(config)

        with self.assertRaisesSystemExit(3):
            self.setup_worker(config)

        fail_mocked.assert_called_with(log_message)

    async def test_invalid_configuration_missing_general_section(self) -> None:
        """Worker() raises an exception: [General] section is not available."""
        config_temp_directory = self.create_temp_config_directory({})
        config = ConfigHandler(directories=[config_temp_directory.as_posix()])

        log_message = (
            f'Missing required section "General" in the configuration file '
            f'({config.active_configuration_file})'
        )

        fail_mocked = self.patch_config_handler_fail(config)

        with self.assertRaisesSystemExit(3):
            self.setup_worker(config)

        fail_mocked.assert_called_with(log_message)

    async def test_invalid_configuration_api_url_not_https(self) -> None:
        """A worker configured to require an HTTPS api-url fails with HTTP."""
        config_temp_directory = self.create_temp_config_directory(
            {"General": {"api-url": "http://example.org/"}}
        )
        config = ConfigHandler(
            directories=[config_temp_directory.as_posix()], require_https=True
        )

        log_message = (
            f'api-url in {config.active_configuration_file} does not use '
            f'HTTPS: http://example.org/'
        )

        fail_mocked = self.patch_config_handler_fail(config)

        with self.assertRaisesSystemExit(3):
            self.setup_worker(config)

        fail_mocked.assert_called_with(log_message)

    async def test_valid_configuration_api_url_https(self) -> None:
        """A worker configured to require an HTTPS api-url accepts HTTPS."""
        config_temp_directory = self.create_temp_config_directory(
            {"General": {"api-url": "https://example.org/"}}
        )
        config = ConfigHandler(
            directories=[config_temp_directory.as_posix()], require_https=True
        )

        self.setup_worker(config)

    def assert_functools_partial_equal(
        self, partial1: functools.partial[Any], partial2: functools.partial[Any]
    ) -> None:
        """Assert two functools.partials are semantically equal."""
        # functools.partial does not implement __eq__
        self.assertEqual(partial1.func, partial2.func)
        self.assertEqual(partial1.args, partial2.args)
        self.assertEqual(partial1.keywords, partial2.keywords)

    async def test_main_calls_connect(self) -> None:
        """Ensure worker.main() calls worker.connect()."""
        self.setup_worker()
        assert self.worker is not None
        mock_connect = self.patch_worker("connect")

        patcher = mock.patch.object(
            self.worker._main_event_loop, "add_signal_handler", autospec=True
        )
        mocked_add_signal_handler = patcher.start()
        self.addCleanup(patcher.stop)

        await self.worker.main()

        sigint_call = mocked_add_signal_handler.mock_calls[0][1]
        self.assertEqual(sigint_call[0], signal.SIGINT)
        self.assert_functools_partial_equal(
            sigint_call[1],
            functools.partial(
                self.worker._create_task_signal_int_term_handler, signal.SIGINT
            ),
        )

        sigterm_call = mocked_add_signal_handler.mock_calls[1][1]
        self.assertEqual(sigterm_call[0], signal.SIGTERM)
        self.assert_functools_partial_equal(
            sigterm_call[1],
            functools.partial(
                self.worker._create_task_signal_int_term_handler, signal.SIGTERM
            ),
        )

        self.assertEqual(len(mocked_add_signal_handler.mock_calls), 2)

        mock_connect.assert_awaited()

    async def test_create_task_signal_int_term_handler(self) -> None:
        """
        Test Task returned from Worker._create_task_signal_int_term_handler.

        Assert that the task logs "Terminated with signal ...".
        Other tests on the behaviour of Worker._signal_int_term_handler
        is implemented in the method test_signal_int_term_handler.

        Testing that Worker._create_task_signal_int_term_handler returns
        a Task that will execute Worker._signal_int_term_handler might be
        possible but only if accessing Task._coro private method.
        """
        self.setup_worker()
        assert self.worker is not None

        test_signals = [signal.SIGINT, signal.SIGTERM]

        for test_signal in test_signals:
            with self.subTest(test_signal=repr(test_signal)):
                sig_handler = self.worker._create_task_signal_int_term_handler(
                    test_signal
                )

                self.assertIsInstance(sig_handler, asyncio.Task)

                with self.assertLogsContains(
                    f"Terminated with signal {test_signal.name}"
                ):
                    await sig_handler

    def assert_logging_setup(
        self,
        argv_setup: dict[str, Any],
        config_setup: dict[str, Any],
        expected: dict[str, Any],
    ) -> None:
        """
        Assert logging setup for argv_setup and config_setup.

        :param argv_setup: parameters that a user might have passed using
          the CLI
        :param config_setup: parameters that might be set in the config.ini
        :param expected: given the argv_setup and config_setup: what is the
          expected configuration
        """
        del self.config['General']['log-file']
        if config_setup['log-file'] is not None:
            self.config['General']['log-file'] = config_setup['log-file']
        if config_setup['log-level'] is not None:
            self.config['General']['log-level'] = logging.getLevelName(
                config_setup['log-level']
            )

        # Normalizes log_level: always using the str representation and not
        # the int
        if argv_setup['log_level'] is not None:
            argv_setup['log_level'] = logging.getLevelName(
                argv_setup['log_level']
            )
        expected['level'] = logging.getLevelName(expected['level'])

        # Patches basicConfig
        patcher = mock.patch(
            'debusine.worker._worker.logging.basicConfig', autospec=True
        )
        mocked_logging = patcher.start()
        self.addCleanup(patcher.stop)

        self.setup_worker(**argv_setup)

        self.assertTrue(mocked_logging.called)
        called_args = mocked_logging.call_args[1]
        self.assertEqual(called_args['filename'], expected['filename'])
        self.assertEqual(called_args['level'], expected['level'])

    def test_setup_logging_no_logging_configuration(self) -> None:
        """Logging system: no explicit setup: use Worker.DEFAULT_LOG_LEVEL."""
        argv_setup = {'log_file': None, 'log_level': None}
        config_setup = {'log-file': None, 'log-level': None}
        expected = {'filename': None, 'level': Worker.DEFAULT_LOG_LEVEL}
        self.assert_logging_setup(argv_setup, config_setup, expected)

    def test_setup_logging_from_constructor(self) -> None:
        """Logging system: use settings from the Worker constructor."""
        log_file = '/var/log/debusine.log'
        argv_setup = {'log_file': log_file, 'log_level': logging.DEBUG}
        config_setup = {'log-file': None, 'log-level': None}
        expected = {'filename': log_file, 'level': logging.DEBUG}

        self.assert_logging_setup(argv_setup, config_setup, expected)

    def test_setup_logging_from_config(self) -> None:
        """Logging system: use settings from the configuration file."""
        log_file = '/var/log/debusine.log'
        argv_setup = {'log_file': None, 'log_level': None}
        config_setup = {'log-file': log_file, 'log-level': logging.DEBUG}
        expected = {'filename': log_file, 'level': logging.DEBUG}
        self.assert_logging_setup(argv_setup, config_setup, expected)

    def test_setup_logging_worker_constructor_overwrites_config(self) -> None:
        """Logging system: use settings from the Worker constructor."""
        log_file = '/var/log/debusine.log'
        argv_setup = {'log_file': log_file, 'log_level': logging.DEBUG}
        config_setup = {'log-file': 'unused', 'log-level': logging.ERROR}
        expected = {'filename': log_file, 'level': logging.DEBUG}

        self.assert_logging_setup(argv_setup, config_setup, expected)

    def test_setup_logging_cannot_open_log_file(self) -> None:
        """Worker raises SystemExit if the log-file cannot be opened."""
        patched = mock.patch('debusine.worker.Worker._fail', autospec=True)

        fail_mocked = patched.start()
        fail_mocked.side_effect = SystemExit(3)
        self.addCleanup(patched.stop)

        with self.assertRaisesRegex(SystemExit, '^3$'):
            self.setup_worker(log_file='/')

        fail_mocked.assert_called()

    def test_log_forced_exit(self) -> None:
        """log_forced_exit logs the event."""
        self.setup_worker()
        assert self.worker is not None

        with self.assertLogsContains(
            "Terminated with signal SIGINT", level=logging.INFO
        ):
            self.worker.log_forced_exit(signal.SIGINT)

    async def test_asyncio_sleep(self) -> None:
        """Worker._asyncio_sleep calls asyncio.sleep."""
        patched = mock.patch('asyncio.sleep', autospec=True)
        mocked = patched.start()
        self.addCleanup(patched.stop)

        await Worker._asyncio_sleep(1.5)

        mocked.assert_called_with(1.5)

    async def test_send_task_result_do_not_send_aborted_task(self) -> None:
        """Assert that _send_task_result does not send the result (aborted)."""
        self.setup_worker()
        assert self.worker is not None
        client = self.worker._create_client()

        work_request_completed_update = (
            self.patch_debusine_work_request_completed_update()
        )

        self.worker._task_name = "Noop"
        self.worker._task_running = Noop({"result": WorkRequestResults.FAILURE})
        self.worker._initialize_statistics()
        self.worker._task_running.abort()

        with self.assertLogsContains(
            "Task: Noop has been aborted", level=logging.INFO
        ):
            await self.worker._send_task_result(
                11, concurrent.futures.Future(), client
            )

        work_request_completed_update.assert_not_called()

    async def test_signal_int_term_handler(self) -> None:
        """Test signal handler logs "Terminated..." and closes the session."""
        self.setup_worker()
        assert self.worker is not None

        await self.worker.connect()
        assert self.worker._aiohttp_client_session is not None

        self.assertFalse(self.worker._aiohttp_client_session.closed)

        with self.assertLogsContains("Terminated with signal SIGINT"):
            await self.worker._signal_int_term_handler(signal.SIGINT)

        self.assertTrue(self.worker._aiohttp_client_session.closed)

    async def test_signal_int_term_handler_cancels_task(self) -> None:
        """Test signal handler aborts() the task."""
        self.setup_worker()
        assert self.worker is not None
        self.worker._task_running = Noop({"result": WorkRequestResults.FAILURE})

        self.assertFalse(self.worker._task_running.aborted)

        with self.assertLogsContains("Terminated with signal SIGTERM"):
            await self.worker._signal_int_term_handler(signal.SIGTERM)

        self.assertTrue(self.worker._task_running.aborted)

    async def test_setup_signal_int_term_handler(self) -> None:
        """Test that SIGINT/SIGTERM setup is done in Worker.main()."""
        self.setup_worker()
        assert self.worker is not None

        self.worker._set_main_event_loop()
        assert self.worker._main_event_loop is not None

        self.assertFalse(
            self.worker._main_event_loop.remove_signal_handler(signal.SIGINT)
        )
        self.assertFalse(
            self.worker._main_event_loop.remove_signal_handler(signal.SIGTERM)
        )

        await self.worker.main()

        self.assertTrue(
            self.worker._main_event_loop.remove_signal_handler(signal.SIGINT)
        )
        self.assertTrue(
            self.worker._main_event_loop.remove_signal_handler(signal.SIGTERM)
        )

    def test_raise_try_again(self) -> None:
        """Assert Worker._raise_try_again() raise tenacity.TryAgain()."""
        self.setup_worker(disable_endless_retry=False)
        assert self.worker is not None

        self.assertRaises(tenacity.TryAgain, self.worker._raise_try_again)

    def test_reset_retry_counter(self) -> None:
        """Assert Worker._reset_retry_counter reset the counter and boolean."""
        self.setup_worker()
        assert self.worker is not None

        retry_state = MagicMock()

        retry_state.attempt_number = 5
        self.worker._reset_retry_attempt_number = True

        self.worker._reset_retry_counter(retry_state)

        self.assertEqual(retry_state.attempt_number, 0)
        self.assertFalse(self.worker._reset_retry_attempt_number)

    async def test_reset_state_cancels_task_statistics_poller(self) -> None:
        """Worker._reset_state makes sure the statistics poller is cancelled."""
        self.setup_worker()
        assert self.worker is not None

        self.worker._task_name = "Noop"
        self.worker._task_running = Noop({"result": WorkRequestResults.FAILURE})
        self.worker._initialize_statistics()
        task_statistics_poller = self.worker._task_statistics_poller
        assert task_statistics_poller is not None
        self.assertFalse(task_statistics_poller.cancelling())

        self.worker._reset_state()

        self.assertTrue(task_statistics_poller.cancelling())
        self.assertIsNone(self.worker._task_statistics_poller)

    async def test_run_after_work_request_hook(self) -> None:
        """Run a simple script via _run_after_work_request_hook."""
        self.setup_worker()
        assert self.worker is not None

        hook_dir = Path(self.create_temporary_directory())
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = hook_dir

        marker = hook_dir / "marker"
        script = hook_dir / "hook"
        script.write_text(
            dedent(
                f"""\
                #!/bin/sh
                touch {marker}
                """
            )
        )
        script.chmod(0o755)
        await self.worker._run_after_work_request_hook()
        self.assertTrue(marker.exists())

    async def test_run_after_work_request_hook_error(self) -> None:
        """Run a simple failing script via _run_after_work_request_hook."""
        self.setup_worker()
        assert self.worker is not None

        hook_dir = Path(self.create_temporary_directory())
        self.worker.AFTER_WORK_REQUEST_HOOK_DIR = hook_dir

        script = hook_dir / "hook"
        script.write_text(
            dedent(
                """\
                #!/bin/sh
                exit 1
                """
            )
        )
        script.chmod(0o755)
        with self.assertLogsContains(
            f"Hooks failed to execute in {str(hook_dir)!r} .\n"
            "stdout:\n"
            "\n"
            "stderr:\n"
            f"run-parts: {str(hook_dir)}/hook exited with return code 1",
            level=logging.WARNING,
        ):
            await self.worker._run_after_work_request_hook()

    async def test_send_task_result_runs_after_work_request_hook(self) -> None:
        """Assert that _send_task_result calls _after_work_request_hook."""
        self.setup_worker()
        assert self.worker is not None

        self.worker._task_name = "Noop"
        self.worker._task_running = Noop({"result": WorkRequestResults.FAILURE})
        self.worker._initialize_statistics()
        self.worker._task_running.abort()
        client = self.worker._create_client()

        with self.assertLogsContains(
            "No hooks in '/nonexistent'", level=logging.DEBUG
        ):
            await self.worker._send_task_result(
                11, concurrent.futures.Future(), client=client
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the base DBTask classes."""

from pathlib import Path
from typing import cast, override
from unittest import mock

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts.models import TaskTypes, WorkRequestResults
from debusine.db.models.tasks import DBTask, DBWorkerProxyTask
from debusine.db.playground import scenarios
from debusine.tasks import BaseTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData
from debusine.tasks.server import TaskDatabaseInterface
from debusine.tasks.tests.helper_mixin import SampleBaseExternalTask
from debusine.test.django import TestCase, preserve_db_task_registry
from debusine.test.test_utils import preserve_task_registry


class DBTaskTests(TestCase):
    """Tests for :py:class:`DBTask`."""

    scenario = scenarios.DefaultContext()

    @preserve_db_task_registry()
    def make_concrete_dbtask(self) -> DBTask[BaseTaskData, BaseDynamicTaskData]:
        """Instantiate a concrete instance of DBTask."""

        class ConcreteDBTask(DBTask[BaseTaskData, BaseDynamicTaskData]):
            """Concrete implementation of DBTask."""

            TASK_TYPE = TaskTypes.WORKER
            TASK_VERSION = 1

            @override
            def build_dynamic_data(
                self, task_database: TaskDatabaseInterface
            ) -> BaseDynamicTaskData:
                return BaseDynamicTaskData()

            @override
            def _execute(self) -> WorkRequestResults:
                raise NotImplementedError()

        work_request = self.playground.create_worker_task(
            task_name="concretedbtask"
        )
        return work_request.get_task()

    def test_is_valid_task_name(self) -> None:
        self.assertTrue(DBTask.is_valid_task_name(TaskTypes.WORKER, "sbuild"))

    def test_is_valid_task_name_case_insensitive(self) -> None:
        self.assertTrue(DBTask.is_valid_task_name(TaskTypes.WORKER, "Sbuild"))

    def test_is_valid_task_name_invalid_type(self) -> None:
        self.assertFalse(
            DBTask.is_valid_task_name(
                cast(TaskTypes, "does-not-exist"), "sbuild"
            )
        )

    def test_is_valid_task_name_invalid(self) -> None:
        self.assertFalse(
            DBTask.is_valid_task_name(TaskTypes.WORKER, "not_a_real_task")
        )

    def test_task_names_worker(self) -> None:
        task_names = DBTask.task_names(TaskTypes.WORKER)
        self.assertGreater(len(task_names), 0)
        self.assertIn("noop", task_names)

    def test_worker_task_names(self) -> None:
        task_names = DBTask.worker_task_names()
        self.assertGreater(len(task_names), 0)
        self.assertIn("noop", task_names)

    def test_is_worker_task(self) -> None:
        self.assertTrue(DBTask.is_worker_task("noop"))

    def test_is_worker_task_false(self) -> None:
        self.assertFalse(DBTask.is_worker_task("nonexistent"))

    def test_class_for_name_invalid_type(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "'not-existing-type' is not a registered task type"
        ):
            DBTask.class_from_name(
                cast(TaskTypes, "not-existing-type"), "non-existing-class"
            )

    def test_class_for_name_sbuild(self) -> None:
        self.assertEqual(
            DBTask.class_from_name(TaskTypes.WORKER, 'sBuIlD').name, "sbuild"
        )

    def test_class_for_name_no_class(self) -> None:
        with self.assertRaisesRegex(
            ValueError,
            "'non-existing-class' is not a registered Worker task_name",
        ):
            DBTask.class_from_name(TaskTypes.WORKER, 'non-existing-class')

    def test_compute_system_required_tags(self) -> None:
        task = self.make_concrete_dbtask()
        self.assertCountEqual(
            task.compute_system_required_tags(),
            {
                wtags.WORKER_TYPE_EXTERNAL,
                wtags.TASK_PREFIX
                + f"{task.TASK_TYPE.lower()}:concretedbtask:version:1",
            },
        )

    def test_compute_system_provided_tags(self) -> None:
        self.playground.create_group("test", [self.scenario.user])
        self.playground.create_group(
            "test", [self.scenario.user], workspace=self.scenario.workspace
        )
        task = self.make_concrete_dbtask()
        ws = self.scenario.workspace
        self.assertCountEqual(
            task.compute_system_provided_tags(),
            [
                ttags.SCOPE_PREFIX + ws.scope.name,
                ttags.WORKSPACE_PREFIX + f"{ws.scope.name}:{ws.name}",
                ttags.GROUP_PREFIX + f"{ws.scope.name}::test",
                ttags.GROUP_PREFIX + f"{ws.scope.name}:{ws.name}:test",
            ],
        )

    def test_compute_user_provided_tags_no_source_package_name(self) -> None:
        task = self.make_concrete_dbtask()
        early_dynamic_task_data = mock.Mock()
        early_dynamic_task_data.get_source_package_name = mock.Mock(
            return_value=None
        )
        tags = task.compute_user_provided_tags(early_dynamic_task_data)
        self.assertCountEqual(tags, [])

    def test_compute_user_provided_tags_with_source_package_name(self) -> None:
        task = self.make_concrete_dbtask()
        early_dynamic_task_data = mock.Mock()
        early_dynamic_task_data.get_source_package_name = mock.Mock(
            return_value="foo"
        )
        tags = task.compute_user_provided_tags(early_dynamic_task_data)
        self.assertCountEqual(tags, [ttags.SOURCE_PACKAGE_PREFIX + "foo"])

    def test_compute_tagsets(self) -> None:
        task = self.make_concrete_dbtask()
        for (
            provided,
            required,
            expected_provided,
            expected_required,
        ) in (
            ({}, {}, [], []),
            (
                {
                    "system": {
                        ttags.GROUP_PREFIX + "group",
                        ttags.SCOPE_PREFIX + "scope",
                        ttags.WORKSPACE_PREFIX + "scope:workspace",
                    },
                },
                {"system": {wtags.CAP_PREFIX + "test"}},
                [
                    ttags.GROUP_PREFIX + "group",
                    ttags.SCOPE_PREFIX + "scope",
                    ttags.WORKSPACE_PREFIX + "scope:workspace",
                ],
                [wtags.CAP_PREFIX + "test"],
            ),
            (
                {
                    "system": {ttags.SOURCE_PACKAGE_PREFIX + "test"},
                    "user": {
                        ttags.GROUP_PREFIX + "group",
                        ttags.SCOPE_PREFIX + "scope",
                        ttags.WORKSPACE_PREFIX + "scope:workspace",
                    },
                },
                {},
                [ttags.SOURCE_PACKAGE_PREFIX + "test"],
                [],
            ),
            (
                {"system": {"a", "c", "b"}, "user": {"d", "c", "e"}},
                {"system": {"f", "e", "d"}},
                ["a", "b", "c", "d", "e"],
                ["d", "e", "f"],
            ),
        ):
            with (
                self.subTest(
                    system_provided=str(provided),
                    system_required=str(required),
                ),
                mock.patch.object(
                    task,
                    "compute_system_provided_tags",
                    return_value=provided.get("system", set()),
                ),
                mock.patch.object(
                    task,
                    "compute_user_provided_tags",
                    return_value=provided.get("user", set()),
                ),
                mock.patch.object(
                    task,
                    "compute_system_required_tags",
                    return_value=required.get("system", set()),
                ),
            ):
                task.apply_task_configuration()
                self.assertEqual(
                    task.work_request.scheduler_tags_provided, expected_provided
                )
                self.assertEqual(
                    task.work_request.scheduler_tags_required, expected_required
                )


class DBProxyTaskTests(TestCase):
    """Tests for :py:class:`DBProxyTask`."""

    @preserve_task_registry()
    @preserve_db_task_registry()
    def test_populate_task_registry(self) -> None:
        # samplebasetask does not exist
        self.assertNotIn(
            "samplebasetask", BaseTask._sub_tasks[TaskTypes.WORKER]
        )
        self.assertNotIn("samplebasetask", DBTask._sub_tasks[TaskTypes.WORKER])

        class SampleBaseTask(
            SampleBaseExternalTask[BaseTaskData, BaseDynamicTaskData],
        ):
            """Sample class to test BaseExternalTask class."""

            @override
            def run(self, execute_directory: Path) -> WorkRequestResults:
                """Unused abstract method from BaseExternalTask."""
                raise NotImplementedError()

        # samplebasetask is registered with BaseTask only
        self.assertIn("samplebasetask", BaseTask._sub_tasks[TaskTypes.WORKER])
        self.assertNotIn("samplebasetask", DBTask._sub_tasks[TaskTypes.WORKER])

        DBWorkerProxyTask.populate_task_registry()

        # samplebasetask is registered with BaseTask and DBTask
        self.assertIn("samplebasetask", BaseTask._sub_tasks[TaskTypes.WORKER])
        self.assertIn("samplebasetask", DBTask._sub_tasks[TaskTypes.WORKER])

    def test_compute_system_required_tags(self) -> None:
        work_request = self.playground.create_worker_task(task_name="noop")
        task = work_request.get_task()
        with (
            mock.patch(
                "debusine.db.models.tasks.DBTask.compute_system_required_tags",
                return_value={"db"},
            ),
            mock.patch(
                "debusine.tasks.BaseTask.compute_system_required_tags",
                return_value={"base"},
            ),
        ):
            self.assertCountEqual(
                task.compute_system_required_tags(), {"db", "base"}
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for work requests."""

from collections import defaultdict
from functools import cached_property

from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest, WorkflowTemplate
from debusine.web.helps import DOCS_BASE_URL
from debusine.web.icons import Icons
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.base import UI


class WorkflowTemplateUI(UI[WorkflowTemplate]):
    """UI helpers for WorkflowTemplate instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this workflow template."""
        return ResourcePlace(
            title=f"Workflow template {self.instance.name}",
            url=self.instance.get_absolute_url(),
            breadcrumb=self.instance.name,
            parent=self.instance.workspace.ui(self.request).place,
        )

    def get_help_url(self) -> str:
        """Return a URL to the appropriate workflow documentation."""
        # We could avoid hardcoding this URL structure by using Sphinx to
        # fetch and parse the intersphinx inventory.  However, Sphinx seems
        # like too heavy a dependency to pull in just for this.
        return (
            f"{DOCS_BASE_URL}/reference/workflows/specs/"
            f"{self.instance.task_name.replace('_', '-')}.html"
        )


class WorkRequestUI(UI[WorkRequest]):
    """UI helpers for WorkRequest instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this work request."""
        is_workflow_root = self.instance == self.instance.workflow_root
        place = ResourcePlace(
            title=f"#{self.instance.id}: {self.instance.get_label()}",
            description=f"Work request #{self.instance.id}:"
            f" {self.instance.get_label()}",
            breadcrumb=(
                f"#{self.instance.id}: {self.instance.get_label()}"
                if is_workflow_root
                else f"#{self.instance.id}"
            ),
            url=self.instance.get_absolute_url(),
            icon=Icons.WORK_REQUEST,
        )
        if self.instance.parent is not None:
            place.parent = self.instance.parent.ui(self.request).place
        elif self.instance.task_type == TaskTypes.WORKFLOW:
            place.parent = self.instance.workspace.ui(
                self.request
            ).place_workflow_list
        else:
            place.parent = self.instance.workspace.ui(self.request).place
        return place

    @cached_property
    def workflow_children(self) -> dict[int, list[WorkRequest]]:
        """Mapping from work request IDs to their children in this workflow."""
        children: defaultdict[int, list[WorkRequest]] = defaultdict(list)
        for wr in (
            WorkRequest.objects.filter(id=self.instance.id)
            .workflow_flattened(include_workflows=True)
            .filter(parent__isnull=False)
            .visible_in_workflow()
            .order_by("id")
            .select_related(
                "workspace__scope",
                "aborted_by",
            )
        ):
            assert wr.parent_id is not None
            children[wr.parent_id].append(wr)
        return children
